import numpy as np
from qtpy import QtGui, QtCore, QtWidgets
from qtpy.QtCore import *
from qtpy.QtGui import *
import sharppy.sharptab as tab
from sharppy.sharptab.constants import *
import datetime
import platform

## routine written by Kelton Halbert
## keltonhalbert@ou.edu

__all__ = ['backgroundText', 'plotText']

class backgroundText(QtWidgets.QFrame):
    '''
    Handles drawing the background frame onto a QPixmap.
    Inherits a QtWidgets.QFrame Object.
    '''
    def __init__(self, **kwargs):
        super(backgroundText, self).__init__()
        self.initUI()

    def initUI(self):
        '''
        Initializes frame variables such as padding,
        width, height, etc, as well as the QPixmap
        that contains the frame drawing.
        '''
        ## set the frame stylesheet
        self.setStyleSheet("QFrame {"
            "  background-color: rgb(0, 0, 0);"
            "  border-width: 1px;"
            "  border-style: solid;"
            "  border-color: #3399CC;}")
        ## set the frame padding
        ## set the height/width variables
        self.lpad = 0; self.rpad = 0
        self.tpad = 5; self.bpad = 0
        self.wid = self.size().width()
        self.hgt = self.size().height()
        self.tlx = self.rpad; self.tly = self.tpad
        self.brx = self.wid; self.bry = self.hgt
        ## do a DPI check to make sure
        ## the text is sized properly!
        fsize = np.floor(.06 * self.hgt)
        self.tpad = np.floor(.03 * self.hgt)
        ## set the font, get the metrics and height of the font
        self.label_font = QtGui.QFont('Helvetica')
        self.label_font.setPixelSize(fsize)
        self.label_metrics = QtGui.QFontMetrics( self.label_font )
        self.label_height = self.label_metrics.xHeight() + self.tpad
        ## the self.ylast variable is used as a running sum for
        ## text placement.
        self.ylast = self.label_height
        ## initialize the QPixmap that will be drawn on.
        self.plotBitMap = QtGui.QPixmap(self.width()-2, self.height()-2)
        self.plotBitMap.fill(self.bg_color)
        ## plot the background frame
        self.plotBackground()
    
    def draw_frame(self, qp):
        '''
        Draws the background frame and the text headers for indices.
        '''
        ## initialize a white pen with thickness 1 and a solid line
        pen = QtGui.QPen(self.fg_color, 1, QtCore.Qt.SolidLine)
        qp.setPen(pen)
        qp.setFont(self.label_font)
        ## set the horizontal grid to be the width of the frame
        ## divided into 8 spaces
        x1 = self.brx / 8
        y1 = 1
        ## draw the header and the indices using a loop.
        ## This loop is a 'horizontal' loop that will plot
        ## the text for a row, keeping the vertical placement constant.
        count = 0
        titles = ['PCL', 'CAPE', 'CINH', 'LCL', 'LI', 'LFC', 'EL']
        for title in titles:
            rect = QtCore.QRect(x1*count, y1, x1*2, self.label_height)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignCenter, title)
            count += 1
        vspace = self.label_height
        if platform.system() == "Windows":
            vspace += self.label_metrics.descent()
        qp.drawLine(0, vspace, self.brx, vspace)
        self.ylast = vspace
    
    def resizeEvent(self, e):
        '''
        Handles when the window gets resized.
        '''
        self.initUI()

    def plotBackground(self):
        '''
        Handles drawing the text background onto
        the QPixmap.
        '''
        ## initialize a QPainter objext
        qp = QtGui.QPainter()
        qp.begin(self.plotBitMap)
        ## draw the frame
        self.draw_frame(qp)
        qp.end()


class plotText(backgroundText):
    updatepcl = Signal(tab.params.Parcel)

    '''
    Handles plotting the indices in the frame.
    Inherits a backgroundText Object that contains
    a QPixmap with the frame drawn on it. All drawing
    gets done on this QPixmap, and then the QPixmap
    gets rendered by the paintEvent function.
    '''
    def __init__(self, pcl_types):
        '''
        Initialize the data from a Profile object passed to 
        this class. It then takes the data it needs from the
        Profile object and converts them into strings that
        can be used to draw the text in the frame.
        
        Parameters
        ----------
        pcl_types :  
        '''
        self.bg_color = QtGui.QColor('#000000')
        self.fg_color = QtGui.QColor('#ffffff')
        self.pw_units = 'in'
        self.temp_units = 'Fahrenheit'

        self.alert_colors = [
            QtGui.QColor('#775000'),
            QtGui.QColor('#996600'),
            QtGui.QColor('#ffffff'),
            QtGui.QColor('#ffff00'),
            QtGui.QColor('#ff0000'),
            QtGui.QColor('#e700df'),
        ]
        self.left_scp_color = QtGui.QColor('#00ffff')

        self.pcl_sel_color = QtGui.QColor('#00ffff')
        self.pcl_cin_hi_color = QtGui.QColor('#993333')
        self.pcl_cin_md_color = QtGui.QColor('#996600')
        self.pcl_cin_lo_color = QtGui.QColor('#00ff00')

        self.use_left = False

        ## get the parcels to be displayed in the GUI
        super(plotText, self).__init__()

        self.prof = None;
        self.pcl_types = pcl_types
        self.parcels = {}
        self.bounds = np.empty((4,2))
        self.setDefaultParcel()

        self.w = SelectParcels(self.pcl_types, self)

    def setDefaultParcel(self):
        idx = np.where(np.asarray(self.pcl_types) == "MU")[0]
        if len(idx) == 0:
            self.skewt_pcl = 0
        else:
            self.skewt_pcl = idx[0]

    def mouseDoubleClickEvent(self, e):
        self.w.show()

    def setParcels(self, prof):
        self.parcels["SFC"] = prof.sfcpcl # Set the surface parcel
        self.parcels["ML"] = prof.mlpcl
        self.parcels["FCST"] = prof.fcstpcl
        self.parcels["MU"] = prof.mupcl
        self.parcels["EFF"] = prof.effpcl
        self.parcels["USER"] = prof.usrpcl

    def setProf(self, prof):
        self.ylast = self.label_height
        self.setParcels(prof)
        self.prof = prof;

        ## either get or calculate the indices, round to the nearest int, and
        ## convert them to strings.
        ## K Index
        self.k_idx = tab.utils.INT2STR( prof.k_idx )
        ## precipitable water
        self.pwat = prof.pwat
        ## 0-3km agl lapse rate
        self.lapserate_3km = tab.utils.FLOAT2STR( prof.lapserate_3km, 1 )
        ## 3-6km agl lapse rate
        self.lapserate_3_6km = tab.utils.FLOAT2STR( prof.lapserate_3_6km, 1 )
        ## 850-500mb lapse rate
        self.lapserate_850_500 = tab.utils.FLOAT2STR( prof.lapserate_850_500, 1 )
        ## 700-500mb lapse rate
        self.lapserate_700_500 = tab.utils.FLOAT2STR( prof.lapserate_700_500, 1 )
        ## convective temperature
        self.convT = prof.convT
        ## sounding forecast surface temperature
        self.maxT = prof.maxT
        #fzl = str(int(self.sfcparcel.hght0c))
        ## 100mb mean mixing ratio
        self.mean_mixr = tab.utils.FLOAT2STR( prof.mean_mixr, 1 )
        ## 150mb mean rh
        self.low_rh = tab.utils.INT2STR( prof.low_rh )
        self.mid_rh = tab.utils.INT2STR( prof.mid_rh )
        ## calculate the totals totals index
        self.totals_totals = tab.utils.INT2STR( prof.totals_totals )
        self.dcape = tab.utils.INT2STR( prof.dcape )
        self.drush = prof.drush
        self.sigsevere = tab.utils.INT2STR( prof.sig_severe )
        self.mmp = tab.utils.FLOAT2STR( prof.mmp, 2 )
        self.esp = tab.utils.FLOAT2STR( prof.esp, 1 )
        self.wndg = tab.utils.FLOAT2STR( prof.wndg, 1 )
        self.tei = tab.utils.INT2STR( prof.tei )

        self.clearData()
        self.plotBackground()
        self.plotData()
        self.update()

    def setPreferences(self, update_gui=True, **prefs):
        self.pw_units = prefs['pw_units']
        self.temp_units = prefs['temp_units']
 
        self.bg_color = QtGui.QColor(prefs['bg_color'])
        self.fg_color = QtGui.QColor(prefs['fg_color'])

        self.pwat_colors = [
            QtGui.QColor(prefs['pwat_m3_color']),
            QtGui.QColor(prefs['pwat_m2_color']),
            QtGui.QColor(prefs['pwat_m1_color']),
            self.fg_color,
            QtGui.QColor(prefs['pwat_p1_color']),
            QtGui.QColor(prefs['pwat_p2_color']),
            QtGui.QColor(prefs['pwat_p3_color']),
        ]

        self.alert_colors = [
            QtGui.QColor(prefs['alert_l1_color']),
            QtGui.QColor(prefs['alert_l2_color']),
            QtGui.QColor(prefs['alert_l3_color']),
            QtGui.QColor(prefs['alert_l4_color']),
            QtGui.QColor(prefs['alert_l5_color']),
            QtGui.QColor(prefs['alert_l6_color']),
        ]
        self.left_scp_color = QtGui.QColor(prefs['alert_lscp_color'])

        self.pcl_sel_color = QtGui.QColor(prefs['pcl_sel_color'])
        self.pcl_cin_hi_color = QtGui.QColor(prefs['pcl_cin_hi_color'])
        self.pcl_cin_md_color = QtGui.QColor(prefs['pcl_cin_md_color'])
        self.pcl_cin_lo_color = QtGui.QColor(prefs['pcl_cin_lo_color'])

        if update_gui:
            self.clearData()
            self.plotBackground()
            self.plotData()
            self.update()

    def setDeviant(self, deviant):
        self.use_left = deviant == 'left'
        
        self.clearData()
        self.plotBackground()
        self.plotData()
        self.update()

    def resizeEvent(self, e):
        '''
        Handles when the window is resized.
        
        Parametes
        ---------
        e: an Event Object
        '''
        super(plotText, self).resizeEvent(e)
        self.plotData()
    
    def paintEvent(self, e):
        '''
        Handles when the window gets painted.
        This renders the QPixmap that the backgroundText
        Object contians. For the actual drawing of the data,
        see the plotData function.
        
        Parametes
        ---------
        e: an Event Object
        
        '''
        super(plotText, self).paintEvent(e)
        qp = QtGui.QPainter()
        qp.begin(self)
        qp.drawPixmap(1, 1, self.plotBitMap)
        qp.end()

    def plotData(self):
        '''
        Handles the drawing of the text onto the QPixmap.
        This is where the actual data gets plotted/drawn.
        '''
        if self.prof is None:
            return

        ## initialize a QPainter object
        qp = QtGui.QPainter()
        qp.begin(self.plotBitMap)
        ## draw the indices
        self.drawConvectiveIndices(qp)
        self.drawIndices(qp)
        self.drawSevere(qp)
        qp.end()

    def clearData(self):
        '''
        Handles the clearing of the pixmap
        in the frame.
        '''
        self.plotBitMap = QtGui.QPixmap(self.width(), self.height())
        self.plotBitMap.fill(self.bg_color)
    
    def drawSevere(self, qp):
        '''
        This handles the severe indices, such as STP, sig hail, etc.
        
        Parameters
        ----------
        qp: QtGui.QPainter object
        
        '''
        ## initialize a pen to draw with.
        pen = QtGui.QPen(QtCore.Qt.yellow, 1, QtCore.Qt.SolidLine)
        self.label_font.setBold(True)
        qp.setFont(self.label_font)
       
        color_list = self.alert_colors
        ## needs to be coded.
        x1 = self.brx / 10
        y1 = self.ylast + self.tpad

        ship = self.prof.ship
        ship_str = tab.utils.FLOAT2STR( ship, 1 )

        if self.use_left:
            stp_fixed = self.prof.left_stp_fixed
            stp_cin = self.prof.left_stp_cin
            scp = self.prof.left_scp
        else:
            stp_fixed = self.prof.right_stp_fixed
            stp_cin = self.prof.right_stp_cin
            scp = self.prof.right_scp

        stp_fixed_str = tab.utils.FLOAT2STR( stp_fixed, 1 )
        stp_cin_str = tab.utils.FLOAT2STR( stp_cin, 1 )
        scp_str = tab.utils.FLOAT2STR( scp, 1 )

        if self.prof.latitude < 0:
            stp_fixed = -stp_fixed
            stp_cin = -stp_cin
            scp = -scp

        # Coloring thresholds provided by Rich Thompson (SPC)
        labels = ['Supercell = ', 'STP (cin) = ', 'STP (fix) = ', 'SHIP = ']
        indices = [scp, stp_cin, stp_fixed, ship]
        index_strs = [scp_str, stp_cin_str, stp_fixed_str, ship_str]
        for label, index_str, index in zip(labels,index_strs,indices):
            rect = QtCore.QRect(x1*7, y1, x1*8, self.label_height)
            if index == '--':
                pen = QtGui.QPen(color_list[0], 1, QtCore.Qt.SolidLine)
            elif label == labels[0]: # STP uses a different color scale
                if index >= 19.95:
                    pen = QtGui.QPen(color_list[5], 1, QtCore.Qt.SolidLine)
                elif index >= 11.95:
                    pen = QtGui.QPen(color_list[4], 1, QtCore.Qt.SolidLine)
                elif index >= 1.95:
                    pen = QtGui.QPen(color_list[2], 1, QtCore.Qt.SolidLine)
                elif index >= .45:
                    pen = QtGui.QPen(color_list[1], 1, QtCore.Qt.SolidLine)
                elif index >= -.45:
                    pen = QtGui.QPen(color_list[0], 1, QtCore.Qt.SolidLine)
                elif index < -.45:
                    pen = QtGui.QPen(self.left_scp_color, 1, QtCore.Qt.SolidLine)
            elif label == labels[1]: # STP effective
                if index >= 8:
                    pen = QtGui.QPen(color_list[5], 1, QtCore.Qt.SolidLine)
                elif index >= 4:
                    pen = QtGui.QPen(color_list[4], 1, QtCore.Qt.SolidLine)
                elif index >= 2:
                    pen = QtGui.QPen(color_list[3], 1, QtCore.Qt.SolidLine)
                elif index >= 1:
                    pen = QtGui.QPen(color_list[2], 1, QtCore.Qt.SolidLine)
                elif index >= .5:
                    pen = QtGui.QPen(color_list[1], 1, QtCore.Qt.SolidLine)
                elif index < .5:
                    pen = QtGui.QPen(color_list[0], 1, QtCore.Qt.SolidLine)
            elif label == labels[2]: # STP fixed
                if index >= 7:
                    pen = QtGui.QPen(color_list[5], 1, QtCore.Qt.SolidLine)
                elif index >= 5:
                    pen = QtGui.QPen(color_list[4], 1, QtCore.Qt.SolidLine)
                elif index >= 2:
                    pen = QtGui.QPen(color_list[3], 1, QtCore.Qt.SolidLine)
                elif index >= 1:
                    pen = QtGui.QPen(color_list[2], 1, QtCore.Qt.SolidLine)
                elif index >= .5:
                    pen = QtGui.QPen(color_list[1], 1, QtCore.Qt.SolidLine)
                else:
                    pen = QtGui.QPen(color_list[0], 1, QtCore.Qt.SolidLine)
            elif label == labels[3]: # SHIP
                if index >= 5:
                    pen = QtGui.QPen(color_list[5], 1, QtCore.Qt.SolidLine)
                elif index >= 2:
                    pen = QtGui.QPen(color_list[4], 1, QtCore.Qt.SolidLine)
                elif index >= 1:
                    pen = QtGui.QPen(color_list[3], 1, QtCore.Qt.SolidLine)
                elif index >= .5:
                    pen = QtGui.QPen(color_list[2], 1, QtCore.Qt.SolidLine)
                else:
                    pen = QtGui.QPen(color_list[0], 1, QtCore.Qt.SolidLine)
            qp.setPen(pen)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignLeft, label + index_str)
            vspace = self.label_height
            if platform.system() == "Windows":
                vspace += self.label_metrics.descent()
            y1 += vspace
        self.label_font.setBold(False)

    def drawIndices(self, qp):
        '''
        Draws the non-parcel indices.
        
        Parameters
        ----------
        qp: QtGui.QPainter object
        
        '''
        qp.setFont(self.label_font)
        ## make the initial x point relatice to the width of the frame.
        x1 = self.brx / 10
        rpad = 5
        tpad = 5

        ## Now we have all the data we could ever want. Time to start drawing
        ## them on the frame.
        ## This starts with the left column.
        std_dev = [ u'(<3\u03C3)', u'(2-3\u03C3)', u'(1-2\u03C3)', '', u'(1-2\u03C3)', u'(2-3\u03C3)', u'(>3\u03C3)' ]
        color = self.pwat_colors[self.prof.pwv_flag + 3] 
        dist_string = std_dev[self.prof.pwv_flag + 3]
        if self.pw_units == 'cm':
            pw_display = tab.utils.IN2CM(self.pwat)
            pw_display = tab.utils.FLOAT2STR( pw_display, 1 )
        else:
            pw_display = self.pwat
            pw_display = tab.utils.FLOAT2STR( pw_display, 2 )

        ## draw the first column of text using a loop, keeping the horizontal
        ## placement constant.
        y1 = self.ylast + self.tpad

        if self.temp_units == 'Fahrenheit':
            t_units_disp = 'F'
            drush_disp = tab.utils.INT2STR( self.drush )
            convT_disp = tab.utils.INT2STR( self.convT )
            maxT_disp = tab.utils.INT2STR( self.maxT )
        elif self.temp_units == 'Celsius':
            t_units_disp = 'C'
            drush_disp = tab.utils.INT2STR( tab.thermo.ftoc(self.drush) )
            convT_disp = tab.utils.INT2STR( tab.thermo.ftoc(self.convT) )
            maxT_disp = tab.utils.INT2STR( tab.thermo.ftoc(self.maxT) )

        colors = [color, self.fg_color, self.fg_color, self.fg_color, self.fg_color, self.fg_color]
        texts = ['PW = ', 'MeanW = ', 'LowRH = ', 'MidRH = ', 'DCAPE = ', 'DownT = ']
        indices = [pw_display + self.pw_units + ' ' + dist_string, self.mean_mixr + 'g/kg', self.low_rh + '%', self.mid_rh + '%', self.dcape, drush_disp + t_units_disp]
        for text, index, c in zip(texts, indices, colors):
            rect = QtCore.QRect(rpad, y1, x1*4, self.label_height)
            pen = QtGui.QPen(c, 1, QtCore.Qt.SolidLine)
            qp.setPen(pen)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignLeft, text + index)
            vspace = self.label_height
            if platform.system() == "Windows":
                vspace += self.label_metrics.descent()
            y1 += vspace

        ## middle-left column
        y1 = self.ylast + self.tpad
        texts = ['K = ', 'TT = ', 'ConvT = ', 'maxT = ', 'ESP = ', 'MMP = ']
        indices = [self.k_idx, self.totals_totals, convT_disp + t_units_disp, maxT_disp + t_units_disp, self.esp, self.mmp]
        for text, index in zip(texts, indices):
            rect = QtCore.QRect(x1*3.5, y1, x1*4, self.label_height)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignLeft, text + index)
            vspace = self.label_height
            if platform.system() == "Windows":
                vspace += self.label_metrics.descent()
            y1 += vspace

        ## middle-right column
        y1 = self.ylast + self.tpad
        texts = ['WNDG = ', 'TEI = ', '3CAPE = ', 'MBURST = ', '', 'SigSvr = ']
        indices = [self.wndg, self.tei, tab.utils.INT2STR(self.prof.mlpcl.b3km), tab.utils.INT2STR(self.prof.mburst), '', self.sigsevere + ' m3/s3']
        for text, index in zip(texts, indices):
            rect = QtCore.QRect(x1*6, y1, x1*4, self.label_height)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignLeft, text + index)
            vspace = self.label_height
            if platform.system() == "Windows":
                vspace += self.label_metrics.descent()
            y1 += vspace
            self.ylast = y1
        qp.drawLine(0, y1+2, self.brx, y1+2)
        qp.drawLine(x1*7-5, y1+2, x1*7-5, self.bry )
        
        ## lapserate window
        y1 = self.ylast + self.tpad
        texts = ['Sfc-3km AGL LR = ', '3-6km AGL LR = ', '850-500mb LR = ', '700-500mb LR = ']
        indices = [self.lapserate_3km + ' C/km', self.lapserate_3_6km + ' C/km', self.lapserate_850_500 + ' C/km', self.lapserate_700_500 + ' C/km']
        for text, index in zip(texts, indices):
            rect = QtCore.QRect(rpad, y1, x1*8, self.label_height)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignLeft, text + index)
            vspace = self.label_height
            if platform.system() == "Windows":
                vspace += self.label_metrics.descent()
            y1 += vspace


    def drawConvectiveIndices(self, qp):
        '''
        This handles the drawing of the parcel indices.
        
        Parameters
        ----------
        qp: QtGui.QPainter object
        
        '''
        ## initialize a white pen with thickness 2 and a solid line
        pen = QtGui.QPen(self.fg_color, 1, QtCore.Qt.SolidLine)
        qp.setPen(pen)
        qp.setFont(self.label_font)
        ## make the initial x pixel coordinate relative to the frame
        ## width.
        x1 = self.brx / 8
        y1 = self.ylast + self.tpad
        ## get the indices rounded to the nearest int, conver to strings
        ## Start with the surface based parcel.
        capes = np.empty(4, dtype="S10") # Only allow 4 parcels to be displayed
        cins = np.empty(4, dtype="S10")
        lcls = np.empty(4, dtype="S10")
        lis = np.empty(4, dtype="S10")
        lfcs = np.empty(4, dtype="S10")
        els = np.empty(4, dtype="S10")
        texts = self.pcl_types
        for i, key in enumerate(texts):
            parcel = self.parcels[key]
            capes[i] = tab.utils.INT2STR(parcel.bplus) # Append the CAPE value
            cins[i] = tab.utils.INT2STR(parcel.bminus)
            lcls[i] = tab.utils.INT2STR(parcel.lclhght)
            lis[i] = tab.utils.INT2STR(parcel.li5)
            lfcs[i] = tab.utils.INT2STR(parcel.lfchght)
            els[i] = tab.utils.INT2STR(parcel.elhght)
        ## Now that we have all the data, time to plot the text in their
        ## respective columns.
        
        ## PCL type
        pcl_index = 0
        for i, text in enumerate(texts):
            self.bounds[i,0] = y1
            if text == self.pcl_types[self.skewt_pcl]:
                pen = QtGui.QPen(self.pcl_sel_color, 1, QtCore.Qt.SolidLine)
                qp.setPen(pen)
                pcl_index = i
            else:
                pen = QtGui.QPen(self.fg_color, 1, QtCore.Qt.SolidLine)
                qp.setPen(pen)
            rect = QtCore.QRect(0, y1, x1*2, self.label_height)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignCenter, text)
            vspace = self.label_height
            if platform.system() == "Windows":
                vspace += self.label_metrics.descent()
            y1 += vspace
            self.bounds[i,1] = y1
        ## CAPE
        y1 = self.ylast + self.tpad
        for i, text in enumerate(capes):
            try:
                if pcl_index == i and int(text) >= 4000:
                    color = self.alert_colors[5]
                elif pcl_index == i and int(text) >= 3000:
                    color = self.alert_colors[4]
                elif pcl_index == i and int(text) >= 2000:
                    color = self.alert_colors[3]
                else:
                    color=self.fg_color
            except:
                color=self.fg_color
            pen = QtGui.QPen(color, 1, QtCore.Qt.SolidLine)
            qp.setPen(pen)
            rect = QtCore.QRect(x1*1, y1, x1*2, self.label_height)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignCenter, bytes(text).decode('utf-8'))
            vspace = self.label_height
            if platform.system() == "Windows":
                vspace += self.label_metrics.descent()
            y1 += vspace
        ## CINH
        y1 = self.ylast + self.tpad
        for i, text in enumerate(cins):
            try:
                if int(capes[i]) > 0 and pcl_index == i and int(text) >= -50:
                    color = self.pcl_cin_lo_color 
                elif int(capes[i]) > 0 and pcl_index == i and int(text) >= -100:
                    color = self.pcl_cin_md_color
                elif int(capes[i]) > 0 and pcl_index == i and int(text) < -100:
                    color = self.pcl_cin_hi_color
                else:
                    color = self.fg_color
            except:
                color = self.fg_color
            pen = QtGui.QPen(color, 1, QtCore.Qt.SolidLine)
            qp.setPen(pen)
            rect = QtCore.QRect(x1*2, y1, x1*2, self.label_height)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignCenter, bytes(text).decode('utf-8'))
            vspace = self.label_height
            if platform.system() == "Windows":
                vspace += self.label_metrics.descent()
            y1 += vspace

        ## LCL
        y1 = self.ylast + self.tpad
        for i, text in enumerate(lcls):
            try:
                if int(text) < 1000 and pcl_index == i and texts[i] == "ML":
                    color = self.pcl_cin_lo_color
                elif int(text) < 1500 and pcl_index == i and texts[i] == "ML":
                    color = self.pcl_cin_md_color
                elif int(text) < 2000 and pcl_index == i and texts[i] == "ML":
                    color = self.pcl_cin_hi_color
                else:
                    color = self.fg_color
            except:
                color=self.fg_color
            pen = QtGui.QPen(color, 1, QtCore.Qt.SolidLine)
            qp.setPen(pen)
            rect = QtCore.QRect(x1*3, y1, x1*2, self.label_height)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignCenter, bytes(text).decode('utf-8'))
            vspace = self.label_height
            if platform.system() == "Windows":
                vspace += self.label_metrics.descent()
            y1 += vspace

        pen = QtGui.QPen(self.fg_color, 1, QtCore.Qt.SolidLine)
        qp.setPen(pen)
        ## LI
        y1 = self.ylast + self.tpad
        for text in lis:
            rect = QtCore.QRect(x1*4, y1, x1*2, self.label_height)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignCenter, bytes(text).decode('utf-8'))
            vspace = self.label_height
            if platform.system() == "Windows":
                vspace += self.label_metrics.descent()
            y1 += vspace
        ## LFC
        y1 = self.ylast + self.tpad
        for text in lfcs:
            rect = QtCore.QRect(x1*5, y1, x1*2, self.label_height)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignCenter, bytes(text).decode('utf-8'))
            vspace = self.label_height
            if platform.system() == "Windows":
                vspace += self.label_metrics.descent()
            y1 += vspace
        ## EL
        y1 = self.ylast + self.tpad
        for text in els:
            rect = QtCore.QRect(x1*6, y1, x1*2, self.label_height)
            qp.drawText(rect, QtCore.Qt.TextDontClip | QtCore.Qt.AlignCenter, bytes(text).decode('utf-8'))
            vspace = self.label_height
            if platform.system() == "Windows":
                vspace += self.label_metrics.descent()
            y1 += vspace
            self.ylast = y1
        qp.drawLine(0, y1+2, self.brx, y1+2)
        color=QtGui.QColor('#996633')
        pen = QtGui.QPen(color, .5, QtCore.Qt.SolidLine)
        qp.setPen(pen)
        qp.drawLine(2, self.bounds[self.skewt_pcl,0]-1, x1*6 + x1*2 - 1, self.bounds[self.skewt_pcl,0]-1)
        qp.drawLine(2, self.bounds[self.skewt_pcl,0]-1, 2, self.bounds[self.skewt_pcl,1])
        qp.drawLine(2, self.bounds[self.skewt_pcl,1], x1*6 + x1*2 - 1, self.bounds[self.skewt_pcl,1])
        qp.drawLine(x1*6 + x1*2 -1 , self.bounds[self.skewt_pcl,0]-1, x1*6 + x1*2 -1, self.bounds[self.skewt_pcl,1])


    def mousePressEvent(self, e):
        pos = e.pos()
        for i, bound in enumerate(self.bounds):
            if bound[0] < pos.y() and bound[1] > pos.y():
                self.skewt_pcl = i
                self.ylast = self.label_height
                pcl_to_pass = self.parcels[self.pcl_types[self.skewt_pcl]]
                self.updatepcl.emit(pcl_to_pass)
                self.clearData()
                self.plotBackground()
                self.plotData()
                self.update()
                self.parentWidget().setFocus()
                break

class SelectParcels(QtWidgets.QWidget):
    def __init__(self, parcel_types, parent):
        QtWidgets.QWidget.__init__(self)
        self.thermo = parent
        self.parcel_types = parcel_types
        self.max_pcls = 4
        self.pcl_count = 0
        self.initUI()

    def initUI(self):

        self.sb = QtWidgets.QCheckBox('Surface-Based Parcel', self)
        self.sb.move(20, 20)
        if "SFC" in self.parcel_types:
            self.sb.toggle()
            self.pcl_count += 1
        self.sb.stateChanged.connect(self.changeParcel)

        self.ml = QtWidgets.QCheckBox('100 mb Mixed Layer Parcel', self)
        self.ml.move(20, 40)
        if "ML" in self.parcel_types:
            self.ml.toggle()
            self.pcl_count += 1
        self.ml.stateChanged.connect(self.changeParcel)

        self.fcst = QtWidgets.QCheckBox('Forecast Surface Parcel', self)
        self.fcst.move(20, 60)
        if "FCST" in self.parcel_types:
            self.fcst.toggle()
            self.pcl_count += 1
        self.fcst.stateChanged.connect(self.changeParcel)

        self.mu = QtWidgets.QCheckBox('Most Unstable Parcel', self)
        self.mu.move(20, 80)
        if "MU" in self.parcel_types:
            self.mu.toggle()
            self.pcl_count += 1
        self.mu.stateChanged.connect(self.changeParcel)

        self.eff = QtWidgets.QCheckBox('Effective Inflow Layer Parcel', self)
        self.eff.move(20, 100)
        if "EFF" in self.parcel_types:
            self.eff.toggle()
            self.pcl_count += 1
        self.eff.stateChanged.connect(self.changeParcel)

        self.usr = QtWidgets.QCheckBox('User Defined Parcel', self)
        self.usr.move(20, 120)
        if "USER" in self.parcel_types:
            self.usr.toggle()
            self.pcl_count += 1
        self.usr.stateChanged.connect(self.changeParcel)


        self.setGeometry(300, 300, 250, 180)
        self.setWindowTitle('Show Parcels')
        self.ok = QtWidgets.QPushButton('Ok', self)
        self.ok.move(20,150)
        self.ok.clicked.connect(self.okPushed)

        #cb.stateChanged.connect(self.changeTitle)
    #cb.stateChanged.connect(

    def changeParcel(self, state):
        if state == QtCore.Qt.Checked:
            self.pcl_count += 1
        else:
            self.pcl_count -= 1

    def okPushed(self):
        if self.pcl_count > self.max_pcls:
            msgBox = QMessageBox()
            msgBox.setText("You can only show 4 parcels at a time.\nUnselect some parcels.")
            msgBox.exec_()
        elif self.pcl_count != self.max_pcls:
            msgBox = QMessageBox()
            msgBox.setText("You need to select 4 parcels to show.  Select some more.")
            msgBox.exec_()
        else:
            self.parcel_types = []
            if self.sb.isChecked() is True:
                self.parcel_types.append('SFC')
            if self.ml.isChecked() is True:
                self.parcel_types.append('ML')
            if self.fcst.isChecked() is True:
                self.parcel_types.append('FCST')
            if self.mu.isChecked() is True:
                self.parcel_types.append('MU')
            if self.eff.isChecked() is True:
                self.parcel_types.append('EFF')
            if self.usr.isChecked() is True:
                self.parcel_types.append('USER')

            self.thermo.pcl_types = self.parcel_types
            self.thermo.skewt_pcl = 0
            self.thermo.ylast = self.thermo.label_height
            pcl_to_pass = self.thermo.parcels[self.thermo.pcl_types[self.thermo.skewt_pcl]]
            self.thermo.updatepcl.emit(pcl_to_pass)
            self.thermo.clearData()
            self.thermo.plotBackground()
            self.thermo.plotData()
            self.thermo.update()
            self.thermo.parentWidget().setFocus()
            self.hide()

if __name__ == '__main__':
    app_frame = QtGui.QApplication([])        
    #tester = plotText(['sfcpcl', 'mlpcl', 'mupcl'])
    tester = plotBackgroundText()
    #tester.setProf()
    tester.show()        
    app_frame.exec_()



''' Wind Manipulation Routines '''
from __future__ import division
import numpy as np
import numpy.ma as ma
from . import interp, utils
from . constants import *
import warnings

__all__ = ['mean_wind', 'mean_wind_npw']
__all__ += ['sr_wind', 'sr_wind_npw', 'wind_shear', 'helicity', 'max_wind']
__all__ += ['non_parcel_bunkers_motion', 'corfidi_mcs_motion', 'mbe_vectors']
__all__ += ['non_parcel_bunkers_motion_experimental', 'critical_angle']

#warnings.warn("Future versions of the routines in the winds module may include options to use height values instead of pressure to specify layers (i.e. SRH, wind shear, etc.)")

def mean_wind(prof, pbot=850, ptop=250, dp=-1, stu=0, stv=0):
    '''
    Calculates a pressure-weighted mean wind through a layer. The default
    layer is 850 to 200 hPa.

    Parameters
    ----------
    prof: profile object
        Profile object
    pbot : number (optional; default 850 hPa)
        Pressure of the bottom level (hPa)
    ptop : number (optional; default 250 hPa)
        Pressure of the top level (hPa)
    dp : negative integer (optional; default -1)
        The pressure increment for the interpolated sounding
    stu : number (optional; default 0)
        U-component of storm-motion vector (kts)
    stv : number (optional; default 0)
        V-component of storm-motion vector (kts)

    Returns
    -------
    mnu : number
        U-component (kts)
    mnv : number
        V-component (kts)

    '''
    if dp > 0: dp = -dp
    if not utils.QC(pbot) or not utils.QC(ptop):
        return ma.masked, ma.masked
    if prof.wdir.count() == 0:
        return ma.masked, ma.masked

    ps = np.arange(pbot, ptop+dp, dp)
    u, v = interp.components(prof, ps)
    # u -= stu; v -= stv
    return ma.average(u, weights=ps)-stu, ma.average(v, weights=ps)-stv


def mean_wind_npw(prof, pbot=850., ptop=250., dp=-1, stu=0, stv=0):
    '''
    Calculates a non-pressure-weighted mean wind through a layer. The default
    layer is 850 to 200 hPa.

    Parameters
    ----------
    prof: profile object
        Profile object
    pbot : number (optional; default 850 hPa)
        Pressure of the bottom level (hPa)
    ptop : number (optional; default 250 hPa)
        Pressure of the top level (hPa)
    dp : negative integer (optional; default -1)
        The pressure increment for the interpolated sounding (mb)
    stu : number (optional; default 0)
        U-component of storm-motion vector (kts)
    stv : number (optional; default 0)
        V-component of storm-motion vector (kts)

    Returns
    -------
    mnu : number
        U-component (kts)
    mnv : number
        V-component (kts)

    '''
    if prof.wdir.count() == 0 or not utils.QC(ptop) or not utils.QC(pbot):
        return ma.masked, ma.masked

    if dp > 0: dp = -dp
    ps = np.arange(pbot, ptop+dp, dp)
    u, v = interp.components(prof, ps)
    # u -= stu; v -= stv
    return u.mean()-stu, v.mean()-stv


def sr_wind(prof, pbot=850, ptop=250, stu=0, stv=0, dp=-1):
    '''
    Calculates a pressure-weighted mean storm-relative wind through a layer.
    The default layer is 850 to 200 hPa. This is a thin wrapper around
    mean_wind().

    Parameters
    ----------
    prof: profile object
        Profile object
    pbot : number (optional; default 850 hPa)
        Pressure of the bottom level (hPa)
    ptop : number (optional; default 250 hPa)
        Pressure of the top level (hPa)
    stu : number (optional; default 0)
        U-component of storm-motion vector (kts)
    stv : number (optional; default 0)
        V-component of storm-motion vector  (kts)
    dp : negative integer (optional; default -1)
        The pressure increment for the interpolated sounding (mb)

    Returns
    -------
    mnu : number
        U-component (kts)
    mnv : number
        V-component (kts)

    '''
    return mean_wind(prof, pbot=pbot, ptop=ptop, dp=dp, stu=stu, stv=stv)


def sr_wind_npw(prof, pbot=850, ptop=250, stu=0, stv=0, dp=-1):
    '''
    Calculates a none-pressure-weighted mean storm-relative wind through a
    layer. The default layer is 850 to 200 hPa. This is a thin wrapper around
    mean_wind_npw().

    Parameters
    ----------
    prof: profile object
        Profile object
    pbot : number (optional; default 850 hPa)
        Pressure of the bottom level (hPa)
    ptop : number (optional; default 250 hPa)
        Pressure of the top level (hPa)
    stu : number (optional; default 0)
        U-component of storm-motion vector (kts)
    stv : number (optional; default 0)
        V-component of storm-motion vector (kts)
    dp : negative integer (optional; default -1)
        The pressure increment for the interpolated sounding (mb)

    Returns
    -------
    mnu : number
        U-component (kts)
    mnv : number
        V-component (kts)

    '''
    return mean_wind_npw(prof, pbot=pbot, ptop=ptop, dp=dp, stu=stu, stv=stv)


def wind_shear(prof, pbot=850, ptop=250):
    '''
    Calculates the shear between the wind at (pbot) and (ptop).

    Parameters
    ----------
    prof: profile object
        Profile object
    pbot : number (optional; default 850 hPa)
        Pressure of the bottom level (hPa)
    ptop : number (optional; default 250 hPa)
        Pressure of the top level (hPa)

    Returns
    -------
    shu : number
        U-component (kts)
    shv : number
        V-component (kts)

    '''
    if prof.wdir.count() == 0 or not utils.QC(ptop) or not utils.QC(pbot):
        return ma.masked, ma.masked

    ubot, vbot = interp.components(prof, pbot)
    utop, vtop = interp.components(prof, ptop)
    shu = utop - ubot
    shv = vtop - vbot
    return shu, shv

def non_parcel_bunkers_motion_experimental(prof):
    '''
        Compute the Bunkers Storm Motion for a Right Moving Supercell
        
        Parameters
        ----------
        prof : profile object
            Profile Object
        
        Returns
        -------
        rstu : number
            Right Storm Motion U-component (kts)
        rstv : number
            Right Storm Motion V-component (kts)
        lstu : number
            Left Storm Motion U-component (kts)
        lstv : number
            Left Storm Motion V-component (kts)
        
        '''
    if prof.wdir.count() == 0:
        return ma.masked, ma.masked, ma.masked, ma.masked

    d = utils.MS2KTS(7.5)     # Deviation value emperically derived as 7.5 m/s
    ## get the msl height of 500m, 5.5km, and 6.0km above the surface
    msl500m = interp.to_msl(prof, 500.)
    msl5500m = interp.to_msl(prof, 5500.)
    msl6000m = interp.to_msl(prof, 6000.)
    
    ## get the pressure of the surface, 500m, 5.5km, and 6.0km levels
    psfc = prof.pres[prof.sfc]
    p500m = interp.pres(prof, msl500m)
    p5500m = interp.pres(prof, msl5500m)
    p6000m = interp.pres(prof, msl6000m)
    
    ## sfc-500m Mean Wind
    mnu500m, mnv500m = mean_wind(prof, psfc, p500m)
    
    ## 5.5km-6.0km Mean Wind
    mnu5500m_6000m, mnv5500m_6000m = mean_wind(prof, p5500m, p6000m)
    
    # shear vector of the two mean winds
    shru = mnu5500m_6000m - mnu500m
    shrv = mnv5500m_6000m - mnv500m
    
    # SFC-6km Mean Wind
    mnu6, mnv6 = mean_wind(prof, psfc, p6000m)
    
    # Bunkers Right Motion
    tmp = d / utils.mag(shru, shrv)
    rstu = mnu6 + (tmp * shrv)
    rstv = mnv6 - (tmp * shru)
    lstu = mnu6 - (tmp * shrv)
    lstv = mnv6 + (tmp * shru)
    
    return rstu, rstv, lstu, lstv


def non_parcel_bunkers_motion(prof):
    '''
    Compute the Bunkers Storm Motion for a Right Moving Supercell

    Parameters
    ----------
    prof : profile object
        Profile Object

    Returns
    -------
    rstu : number
        Right Storm Motion U-component (kts)
    rstv : number
        Right Storm Motion V-component (kts)
    lstu : number
        Left Storm Motion U-component (kts)
    lstv : number
        Left Storm Motion V-component (kts)

    '''
    if prof.wdir.count() == 0:
        return ma.masked, ma.masked, ma.masked, ma.masked

    d = utils.MS2KTS(7.5)     # Deviation value emperically derived as 7.5 m/s
    msl6km = interp.to_msl(prof, 6000.)
    p6km = interp.pres(prof, msl6km)
    # SFC-6km Mean Wind
    mnu6, mnv6 = mean_wind_npw(prof, prof.pres[prof.sfc], p6km)
    # SFC-6km Shear Vector
    shru, shrv = wind_shear(prof, prof.pres[prof.sfc], p6km)

    # Bunkers Right Motion
    tmp = d / utils.mag(shru, shrv)
    rstu = mnu6 + (tmp * shrv)
    rstv = mnv6 - (tmp * shru)
    lstu = mnu6 - (tmp * shrv)
    lstv = mnv6 + (tmp * shru)

    return rstu, rstv, lstu, lstv


def helicity(prof, lower, upper, stu=0, stv=0, dp=-1, exact=True):
    '''
    Calculates the relative helicity (m2/s2) of a layer from lower to upper.
    If storm-motion vector is supplied, storm-relative helicity, both
    positve and negative, is returned.

    Parameters
    ----------
    prof : profile object
        Profile Object
    lower : number
        Bottom level of layer (m, AGL)
    upper : number
        Top level of layer (m, AGL)
    stu : number (optional; default = 0)
        U-component of storm-motion (kts)
    stv : number (optional; default = 0)
        V-component of storm-motion (kts)
    dp : negative integer (optional; default -1)
        The pressure increment for the interpolated sounding (mb)
    exact : bool (optional; default = True)
        Switch to choose between using the exact data (slower) or using
        interpolated sounding at 'dp' pressure levels (faster)

    Returns
    -------
    phel+nhel : number
        Combined Helicity (m2/s2)
    phel : number
        Positive Helicity (m2/s2)
    nhel : number
        Negative Helicity (m2/s2)

    '''
    if prof.wdir.count() == 0 or not utils.QC(lower) or not utils.QC(upper) or not utils.QC(stu) or not utils.QC(stv):
        return ma.masked, ma.masked, ma.masked

    if lower != upper:
        lower = interp.to_msl(prof, lower)
        upper = interp.to_msl(prof, upper)
        plower = interp.pres(prof, lower)
        pupper = interp.pres(prof, upper)
        if np.isnan(plower) or np.isnan(pupper) or \
            type(plower) == type(ma.masked) or type(pupper) == type(ma.masked):
            return np.ma.masked, np.ma.masked, np.ma.masked
        if exact:
            ind1 = np.where(plower >= prof.pres)[0].min()
            ind2 = np.where(pupper <= prof.pres)[0].max()
            u1, v1 = interp.components(prof, plower)
            u2, v2 = interp.components(prof, pupper)
            u = np.concatenate([[u1], prof.u[ind1:ind2+1].compressed(), [u2]])
            v = np.concatenate([[v1], prof.v[ind1:ind2+1].compressed(), [v2]])
        else:
            ps = np.arange(plower, pupper+dp, dp)
            u, v = interp.components(prof, ps)
        sru = utils.KTS2MS(u - stu)
        srv = utils.KTS2MS(v - stv)
        layers = (sru[1:] * srv[:-1]) - (sru[:-1] * srv[1:])
        phel = layers[layers > 0].sum()
        nhel = layers[layers < 0].sum()
    else:
        phel = nhel = 0

    return phel+nhel, phel, nhel


def max_wind(prof, lower, upper, all=False):
    '''
    Finds the maximum wind speed of the layer given by lower and upper levels.
    In the event of the maximum wind speed occurring at multiple levels, the
    lowest level it occurs is returned by default.

    Parameters
    ----------
    prof : profile object
        Profile Object
    lower : number
        Bottom level of layer (m, AGL)
    upper : number
        Top level of layer (m, AGL)
    all : Boolean
        Switch to change the output to sorted wind levels or maximum level.

    Returns
    -------
    p : number, numpy array
        Pressure level (hPa) of max wind speed
    maxu : number, numpy array
        Maximum Wind Speed U-component (kts)
    maxv : number, numpy array
        Maximum Wind Speed V-component (kts)

    '''
    if prof.wdir.count() == 0 or not utils.QC(lower) or not utils.QC(upper):
        return ma.masked, ma.masked, ma.masked

    lower = interp.to_msl(prof, lower)
    upper = interp.to_msl(prof, upper)
    plower = interp.pres(prof, lower)
    pupper = interp.pres(prof, upper)
    if np.ma.is_masked(plower) or np.ma.is_masked(pupper):
        warnings.warn("winds.max_wind() was unable to interpolate between height and pressure correctly.  This may be due to a data integrity issue.")
        return ma.masked, ma.masked, ma.masked
    #print(lower, upper, plower, pupper, prof.pres)
    ind1 = np.where((plower > prof.pres) | (np.isclose(plower, prof.pres)))[0][0]
    ind2 = np.where((pupper < prof.pres) | (np.isclose(pupper, prof.pres)))[0][-1]

    if len(prof.wspd[ind1:ind2+1]) == 0 or ind1 == ind2:
        maxu, maxv =  utils.vec2comp([prof.wdir[ind1]], [prof.wspd[ind1]])
        return maxu, maxv, prof.pres[ind1]

    arr = prof.wspd[ind1:ind2+1]
    inds = np.ma.argsort(arr)
    inds = inds[~arr[inds].mask][::-1]
    maxu, maxv =  utils.vec2comp(prof.wdir[ind1:ind2+1][inds], prof.wspd[ind1:ind2+1][inds])
    if all:
        return maxu, maxv, prof.pres[inds]
    else:
        return maxu[0], maxv[0], prof.pres[inds[0]]


def corfidi_mcs_motion(prof):
    '''
    Calculated the Meso-beta Elements (Corfidi) Vectors

    Parameters
    ----------
    prof : profile object
        Profile Object

    Returns
    -------
    upu : number
        U-component of the upshear vector (kts)
    upv : number
        V-component of the upshear vector (kts)
    dnu : number
        U-component of the downshear vector (kts)
    dnv : number
        V-component of the downshear vector (kts)

    '''
    if prof.wdir.count() == 0:
        return ma.masked, ma.masked, ma.masked, ma.masked
    # Compute the tropospheric (850hPa-300hPa) mean wind
    if prof.pres[ prof.sfc ] < 850:
         mnu1, mnv1 = mean_wind_npw(prof, pbot=prof.pres[prof.sfc], ptop=300.)
    else:
        mnu1, mnv1 = mean_wind_npw(prof, pbot=850., ptop=300.)

    # Compute the low-level (SFC-1500m) mean wind
    p_1p5km = interp.pres(prof, interp.to_msl(prof, 1500.))
    mnu2, mnv2 = mean_wind_npw(prof, prof.pres[prof.sfc], p_1p5km)

    # Compute the upshear vector
    upu = mnu1 - mnu2
    upv = mnv1 - mnv2

    # Compute the downshear vector
    dnu = mnu1 + upu
    dnv = mnv1 + upv

    return upu, upv, dnu, dnv


def mbe_vectors(prof):
    '''
    Thin wrapper around corfidi_mcs_motion()

    Parameters
    ----------
    prof : profile object
        Profile Object

    Returns
    -------
    upu : number
        U-component of the upshear vector (kts)
    upv : number
        V-component of the upshear vector (kts)
    dnu : number
        U-component of the downshear vector (kts)
    dnv : number
        V-component of the downshear vector (kts)

    '''
    return corfidi_mcs_motion(prof)

def critical_angle(prof, stu=0, stv=0):
    '''
    Calculates the critical angle (degrees) as specified by Esterheld and Giuliano (2008).
    If the critical angle is 90 degrees, this indicates that the lowest 500 meters of 
    the storm is experiencing pure streamwise vorticity.

    Parameters
    ----------
    prof : profile object
        Profile Object
    stu : number (optional; default = 0)
        U-component of storm-motion (kts)
    stv : number (optional; default = 0)
        V-component of storm-motion (kts)

    Returns
    -------
    angle : number
        Critical Angle (degrees)

    '''
    if prof.wdir.count() == 0:
        return ma.masked

    if not utils.QC(stu) or not utils.QC(stv):
        return ma.masked

    pres_500m = interp.pres(prof, interp.to_msl(prof, 500))
    u500, v500 = interp.components(prof, pres_500m)
    sfc_u, sfc_v = interp.components(prof, prof.pres[prof.sfc])

    vec1_u, vec1_v = u500 - sfc_u, v500 - sfc_v    
    vec2_u, vec2_v = stu - sfc_u, stv - sfc_v
    vec_1_mag = np.sqrt(np.power(vec1_u, 2) + np.power(vec1_v, 2))
    vec_2_mag = np.sqrt(np.power(vec2_u, 2) + np.power(vec2_v, 2))

    dot = vec1_u * vec2_u + vec1_v * vec2_v
    angle = np.degrees(np.arccos(dot / (vec_1_mag * vec_2_mag)))

    return angle




import numpy as np

from ... sutils.utils import is_py3

try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen
import certifi
from datetime import datetime
import glob
import os
import imp
import logging

class abstract(object):
    def __init__(self, func):
        self._func = func

    def __call__(self, *args, **kwargs):
        raise NotImplementedError("Function or method '%s' is abstract.  Override it in a subclass!" % self._func.__name__)

# Comment this file

HOME_DIR = os.path.join(os.path.expanduser("~"), ".sharppy", "decoders")
_decoders = {}

def findDecoders():
    global _decoders

    level = -1 if not is_py3() else 0

    built_ins = [ 'buf_decoder', 'spc_decoder', 'pecan_decoder', 'arw_decoder', 'uwyo_decoder', 'nucaps_decoder' ] # JTS - Added NUCAPS.
    io = __import__('sharppy.io', globals(), locals(), built_ins, level)

    for dec in built_ins:
        # Load build-in decoders
        logging.debug("Loading decoder '%s'." % dec)
        dec_imp = getattr(io, dec)

        dec_name = dec_imp.__classname__
        fmt_name = dec_imp.__fmtname__

        _decoders[fmt_name] = getattr(dec_imp, dec_name)

    custom = glob.glob(os.path.join(HOME_DIR, '*.py'))

    for dec in custom:
        # Find and load custom decoders
        dec_mod_name = os.path.basename(dec)[:-3]
        logging.debug("Found custom decoder '%s'." % dec_mod_name)
        dec_imp = imp.load_source(dec_mod_name, dec)

        dec_name = dec_imp.__classname__
        fmt_name = dec_imp.__fmtname__

        _decoders[fmt_name] = getattr(dec_imp, dec_name)

def getDecoder(dec_name):
    return getDecoders()[dec_name]

def getDecoders():
    if _decoders == {}:
        findDecoders()

    return _decoders

class Decoder(object):
    def __init__(self, file_name):
        self._file_name = file_name
        self._prof_collection = self._parse()

    @abstract
    def _parse(self):
        pass

    def _downloadFile(self):
        # Try to open the file.  This is a dirty hack right now until
        # I can figure out a cleaner way to make sure the file (either local or URL)
        # gets opened.
        try:
            f = urlopen(self._file_name, cafile=certifi.where())
        except (ValueError, IOError):
            try:
                fname = self._file_name[7:] if self._file_name.startswith('file://') else self._file_name
                f = open(fname, 'rb')
            except IOError:
                raise IOError("File '%s' cannot be found" % self._file_name)
        file_data = f.read()
#       f.close() # Apparently, this multiplies the time this function takes by anywhere from 2 to 6 ... ???
        return file_data.decode('utf-8')

    def getProfiles(self, indexes=None):
        '''
            Returns a list of profile objects generated from the
            file that was read in.

            Parameters
            ----------
            prof_idxs : list (optional)
                A list of indices corresponding to the profiles to be returned.
                Default is to return the full list of profiles

        '''
        prof_col = self._prof_collection
        if indexes is not None:
            prof_col = prof_col.subset(indexes)
        return prof_col

    def getStnId(self):
        return self._prof_collection.getMeta('loc')

if __name__ == "__main__":
    print("Creating bufkit decoder ...")
    bd = BufDecoder()



import sharppy.databases.sars_cal as sars_cal

def test_sars_hail():
    results = sars_cal.check_hail_cal()
    verif = sars_cal.calc_verification(results)
    # Ensure every sounding listed in the database matches back to itself
    assert verif['num'] == verif['match']

def test_sars_supercell():
    results = sars_cal.check_supercell_cal()
    verif = sars_cal.calc_verification(results)
    # Ensure every sounding listed in the database matches back to itself
    assert verif['num'] == verif['match']

#test_sars_hail()



__author__ = 'Kelton Halbert'

__all__ = ['backgroundGeneric', 'plotGeneric']

import numpy as np
from sharppy.sharptab.constants import *
import sharppy.sharptab as tab
from qtpy import QtGui, QtCore, QtWidgets
from qtpy.QtGui import *
from qtpy.QtCore import *
from qtpy.QtOpenGL import *

class backgroundGeneric(QtWidgets.QFrame):
    """
    A generic class for drawing the background of a widget.
    """

    def __init__(self, **kwargs):
        super(backgroundGeneric, self).__init__()

        ## set the xtick and ytick ranges
        self.xticks = kwargs.get('xticks', None)
        self.yticks = kwargs.get('yticks', None)

        ## initialize background and border color
        self.backgroundColor = kwargs.get('background', BLACK)
        self.borderColor = kwargs.get('border', WHITE)

        self.label_font = QtGui.QFont('Helvetica', 7)

        self.xmin = 0; self.xmax = 100
        self.ymin = 0; self.ymax=100
        self.title = kwargs.get("title", None)
        self.initUI()

    def initUI(self):
        """
        Initialize the user interface. This will set necessary constants
        such as margins, min/max axes values, font sizes, etc.

        :return: None
        """

        ## initialize the frame variables
        self.lpad = 0.; self.rpad = 0.
        self.bpad = 20.; self.tpad = 0.
        self.wid = self.size().width() - self.rpad
        self.hgt = self.size().height() - self.bpad
        self.tlx = self.rpad; self.tly = self.tpad
        self.brx = self.wid; self.bry = self.hgt

        ## initialize the QPixmap that everything gets drawn on
        self.plotBitMap = QtGui.QPixmap(self.size().width(), self.size().height())
        self.plotBitMap.fill(QtGui.QColor(self.backgroundColor))

    def plotBackground(self):
        """
        Handles the drawing of the background on the widget.
        :return: None
        """

        ## initialize a new QPainter object for drawing the stuffs
        qp = QtGui.QPainter()
        ## draw onto the QPixmap
        qp.begin(self.plotBitMap)
        ## these settings usually give better performance
        qp.setRenderHint(qp.Antialiasing)
        qp.setRenderHint(qp.TextAntialiasing)
        ## draw background frame
        self.draw_frame(qp)
        ## draw the x and y tick axes
        self.draw_xticks(qp)
        self.draw_yticks(qp)
        qp.end()
        if self.title is not None:
            self.draw_text(self.xmin + (self.xmax - self.xmin)/2, self.ymax - (self.ymax / 3), self.title)

    def resizeEvent(self, e):
        """
        Handles what to do when the widget is resized.
        This will call self.initUI() when resized in order
        to re-generate the necessary variables.

        :return: None
        """

        self.initUI()

    def x_to_pix(self, x):
        """
        Scale an x coordinate value to pixel space and return it.

        :param x: An x coordinate value
        :return: x converted to pixel space
        """

        scl1 = self.xmax - self.xmin
        scl2 = self.xmax - x
        return self.bry - (scl2 / scl1) * (self.bry - self.tpad)

    def y_to_pix(self, y):
        """
        Scale a y coordinate value to pixel space and return it.

        :param y: A y coordinate value
        :return: y converted to pixel space
        """

        scl1 = self.ymax - self.ymin
        scl2 = self.ymax - y
        return self.bry - (scl2 / scl1 ) * ( self.bry - self.tpad)

    def draw_frame(self, qp):
        """
        Draws the background frame for the widget. The background
        frame includes the background color and the frame border.
        This is primarily used in the initUI function and is not
        necessarily intended to be called outside of it.

        :param qp: QtGui.QPainter object
        :return: None
        """

        ## initialize a new pen and brush
        pen = QtGui.QPen(QtGui.QColor(self.borderColor), 2, QtCore.Qt.SolidLine)
        qp.setPen(pen)
        ## draw the borders in white
        qp.drawLine(self.tlx, self.tly, self.brx, self.tly)
        qp.drawLine(self.brx, self.tly, self.brx, self.bry)
        qp.drawLine(self.brx, self.bry, self.tlx, self.bry)
        qp.drawLine(self.tlx, self.bry, self.tlx, self.tly)

    def draw_text(self, x, y, text, fontsize=12, color=WHITE, linewidth=1, font="Helvetica"):
        """
        Draw text on the widget.
        :return:
        """
        ## initialize a painter and paint on the bitmap
        qp = QtGui.QPainter()
        qp.begin(self.plotBitMap)
        qp.setRenderHint(qp.Antialiasing)
        qp.setRenderHint(qp.TextAntialiasing)

        ## initialize a pen
        text_font = QtGui.QFont(font, fontsize)
        pen = QtGui.QPen(QtGui.QColor(color), linewidth, QtCore.Qt.SolidLine)
        qp.setPen(pen)
        qp.setFont(text_font)

        ## convert the x,y values to pixel space
        xx = self.x_to_pix(x)
        yy = self.y_to_pix(y)
        rect0 = QtCore.QRect(xx, yy, self.brx, self.bry)
        ##qp.drawText(rect0, QtCore.Qt.TextDontClip | QtCore.Qt.AlignCenter, text)
        qp.drawText(xx, yy, 0, 0, QtCore.Qt.TextDontClip | QtCore.Qt.AlignCenter, text)
        qp.end()

    def draw_xticks(self, qp):
        """
        Dray the xtick labels for the frame.

        :param qp: a QtGui.QPainter object
        :return: None
        """
        ## if there are no specified ticks, don't draw them.
        if self.xticks is None:
            return

        ## initialize a pen with the same color as the border
        pen = QtGui.QPen(QtGui.QColor(self.borderColor), 2, QtCore.Qt.SolidLine)
        qp.setPen(pen)
        ## set the font
        qp.setFont(self.label_font)

        ticks = self.xticks
        for tick in ticks:
            ## convert to pixel space
            x = self.x_to_pix(tick)
            ## draw a line of 5 pc as the tick
            qp.drawLine(x, self.bry, x, self.bry - 5.)
            ## draw the text for the tick
            qp.drawText(x, self.bry - 10., 0, 0,
                QtCore.Qt.TextDontClip | QtCore.Qt.AlignCenter,
                tab.utils.INT2STR(tick))

    def draw_yticks(self, qp):
        """
        Dray the ytick labels for the frame.

        :param qp: a QtGui.QPainter object
        :return: None
        """
        ## if there are no ticks, don't draw them.
        if self.yticks is None:
            return

        ## initialize a pen with the same color as teh border
        pen = QtGui.QPen(QtGui.QColor(self.borderColor), 2, QtCore.Qt.SolidLine)
        qp.setPen(pen)
        ## set the font
        qp.setFont(self.label_font)

        ## loop over the ticks
        ticks = self.yticks
        for tick in ticks:
            ## convert to pixel space
            y = self.y_to_pix(tick)
            ## draw a line of length 5px as the tick
            qp.drawLine(0, y, 5, y)
            ## draw the text for the tick
            qp.drawText(10, y, 0, 0,
                QtCore.Qt.TextDontClip | QtCore.Qt.AlignVCenter,
                tab.utils.INT2STR(tick))


    def set_xlim(self, xmin, xmax):
        """
        Set the range of values for the x axis.

        :param xmin: The minimum x axis value
        :param xmax: The maximum x axis value
        :return: None
        """

        self.xmin = xmin; self.xmax = xmax

    def set_ylim(self, ymin, ymax):
        """
        Set the range of values for the y axis.

        :param ymin: The minimum y axis value
        :param ymax: The maximum y axis value
        :return: None
        """

        self.ymin = ymin; self.ymax = ymax


class plotGeneric(backgroundGeneric):
    """
    A generic object to plot the foreground of a plot.
    This inherits from the Background Class that holds
    frame constants, getters, and setters.
    """
    def __init__(self, x, y, **kwargs):
        ## construct the super object
        super(plotGeneric, self).__init__(**kwargs)
        ## set some object variables
        self.x = x; self.y = y
        self.color = kwargs.get('color', RED)
        self.width = kwargs.get('width', 2)
        self.xaxislimit = kwargs.get('xlim', (self.x.min()-10, self.x.max()+10))
        self.yaxislimit = kwargs.get('ylim', (self.y.min(), self.y.max()))
        ## update the min and max bounds based on the data
        self.set_xlim( self.xaxislimit[0], self.xaxislimit[1])
        self.set_ylim( self.yaxislimit[0], self.yaxislimit[1])
        ## reset the x and y limits based on the data given

    def setProf(self, x, y, **kwargs):
        self.x = x; self.y = y
        self.color = kwargs.get('color', RED)
        self.width = kwargs.get('width', 2)
        self.xaxislimit = kwargs.get('xlim', (self.x.min()-10, self.x.max()+10))
        self.yaxislimit = kwargs.get('ylim', (self.y.min(), self.y.max()))
        ## update the min and max bounds based on the data
        self.set_xlim( self.xaxislimit[0], self.xaxislimit[1])
        self.set_ylim( self.yaxislimit[0], self.yaxislimit[1])
        ## reset the x and y limits based on the data given

        self.clearData()
        self.plotBackground()
        self.plotData()
        self.update()

    def resizeEvent(self, e):
        """
        Handles the resizing of the frame

        :param e: an Event object
        :return: None
        """
        super(plotGeneric, self).resizeEvent(e)
        ## you have to call update after setting the variables
        self.plotBackground()
        self.plotData()

    def paintEvent(self, e):
        """
        Handles the paint event and calls the functions
        to draw on the frame.

        :param e: an Event object
        :return: None
        """

        super(backgroundGeneric, self).paintEvent(e)
        qp = QtGui.QPainter()
        qp.begin(self)
        qp.drawPixmap(0, 0, self.plotBitMap)
        qp.end()

    def plotData(self):
        """
        Plot the data on the frame.
        :return: None
        """
        ## create a new painter object
        qp = QtGui.QPainter()
        qp.begin(self.plotBitMap)
        qp.setRenderHint(qp.Antialiasing)
        qp.setRenderHint(qp.TextAntialiasing)
        self.draw_lines(qp)
        ## end the painter
        qp.end()

    def clearData(self):
        '''
        Handles the clearing of the pixmap
        in the frame.
        '''
        self.plotBitMap = QtGui.QPixmap(self.size().width(), self.size().height())
        self.plotBitMap.fill(QtCore.Qt.black)

    def draw_lines(self, qp):
        """
        Draw the line profile on the widget

        :param qp: QtGui.QPainter object
        :return: None
        """
        ## initialize a new pen
        pen = QtGui.QPen(QtGui.QColor(self.color), self.width)
        pen.setStyle(QtCore.Qt.SolidLine)
        qp.setPen(pen)
        ## initialize a painter path
        path = QPainterPath()

        ## get the data mask if there is one
        try:
            mask1 = self.x.mask
            mask2 = self.y.mask
            mask = np.maximum(mask1, mask2)
            y = self.y[~mask]
            x = self.x[~mask]

        ## if the data is not masked, just use the provided arrays!
        except:
            y = self.y
            x = self.x

        ## start the path at the first data value
        path.moveTo(self.x_to_pix(x[0]), self.y_to_pix(y[0]))
        ## now we need to loop through the array
        for i in range( 1, y.shape[0] ):
            ## make sure we are plotting in our minimum and maximum bounds
            if y[i] > self.ymin and y[i] < self.ymax:
                xp = x[i]; yp = y[i]
                ## convert from unit space to pixel space
                x1 = self.x_to_pix(xp)
                y1 = self.y_to_pix(yp)
                ## set the pen and draw a line between the two points
                path.lineTo(x1, y1)

            ## if we are outside our bounds, stop drawing
            else:
                continue
        qp.drawPath(path)

if __name__ == '__main__':
    app_frame = QtGui.QApplication([])        
    tester = plotGeneric()
    #tester.setProf()
    tester.show()        
    app_frame.exec_()



from .skew import plotSkewT
from .hodo import plotHodo
from .thermo import plotText
from .thetae import plotThetae
from .srwinds import plotWinds
from .speed import plotSpeed
from .kinematics import plotKinematics
from .analogues import plotAnalogues
from .watch import plotWatch
from .slinky import plotSlinky
from .advection import plotAdvection
from .stp import plotSTP
from .winter import plotWinter
from .fire import plotFire
from .ship import plotSHIP
from .stpef import plotSTPEF
from .vrot import plotVROT
from .generic import plotGeneric
from .ensemble import plotENS
from .map import Mapper, MapWidget
__all__ = ['plotSkewT', 'plotHodo', 'plotText', 'plotThetae', 'plotWinds', 'plotSpeed',\
           'plotKinematics', 'plotAnalogues', 'plotWatch', 'plotSlinky', 'plotAdvection',\
           'plotSTP', 'plotWinter', 'plotFire', 'plotSHIP', 'plotSTPEF', 'plotVROT',\
           'Mapper', 'MapWidget', 'plotGeneric', 'plotENS']




import numpy as np

from qtpy.QtGui import *
from qtpy.QtCore import *

class Draggable(object):
    """
    The Draggable class implements low-level clicking and dragging on widgets.
    """
    def __init__(self, x_obj, y_obj, background, cutoff=5, lock_dim=None, line_color=QColor('#FFFFFF')):
        """
        Construct a draggable object
        x_obj:  Numpy array or Python float containing the x coordinate(s) in pixels of the draggable object.
        y_obj:  Numpy array or Python float containing the y coordinate(s) in pixels of the draggable object.
        background: A QPixmap containing the image displayed in the widget.
        cutoff: The radius of the buffer around each vertex when looking for clicks. Default is 5 px.
        lock_dim: A dimension to lock in place when doing the dragging. Allowed values are 'x' (dragging
                    is allowed in the y dimension only), 'y' (dragging is allowed in the x direction 
                    only). Default is to allow dragging in both dimensions.
        line_color: The color to use for the line drawn while dragging.
        """
        self._background = background
        self._x_obj = x_obj
        self._y_obj = y_obj
        self._cutoff = cutoff
        self._lock_dim = lock_dim
        self._line_color = line_color
        self._drag_idx = None
        self._save_bitmap = None
        self._drag_buffer = 5

        self._click_start = None

        if type(self._x_obj) != np.ma.MaskedArray:
            self._x_obj = np.ma.array([ self._x_obj ])
        if type(self._y_obj) != np.ma.MaskedArray:
            self._y_obj = np.ma.array([ self._y_obj ])

    def click(self, click_x, click_y):
        """
        Check to see whether the user clicked on a vertex in this profile. If they did, initialize
            the drag operation.
        click_x:    x coordinate in pixels of the click
        click_y:    y coordinate in pixels of the click
        Returns a boolean specifying whether a drag operation was initialized.
        """
        dists = np.hypot(click_x - self._x_obj, click_y - self._y_obj)
        if np.any(dists <= self._cutoff):
            self._drag_idx = np.argmin(dists)
            self._click_start = (click_x, click_y)

        return self._drag_idx is not None

    def drag(self, drag_x, drag_y, restrictions=None):
        """
        Drag a point or a vertex in a line. Applies restrictions to the drag point based on the lock_dim
            argument to the constructor and the restrictions argument to this method.
        drag_x: x coordinate in pixels of the point to drag to.
        drag_y: y coordinate in pixels of the point to drag to.
        restrictions:   A callable object (e.g. callback function) that takes two arguments (x and y 
                            coordinates) and returns a modified x and y based on restrictions on the
                            dragging. Used, for example, to restrict dragging on the temperature line
                            to be greater than the dewpoint.
        """
        if self._drag_idx is None:
            return

        if self._lock_dim == 'x':
            drag_x = self._x_obj[self._drag_idx]
        elif self._lock_dim == 'y':
            drag_y = self._y_obj[self._drag_idx]

        if restrictions is not None:
            drag_x, drag_y = restrictions(drag_x, drag_y)

        if len(self._x_obj) == 1:
            self._dragPoint(drag_x, drag_y)
        else:
            self._dragLine(drag_x, drag_y)

    def _dragLine(self, drag_x, drag_y):
        """
        Drag a line [private method].
        drag_x: x coordinate in pixels of the drag point.
        drag_y: y coordinate in pixels of the drag point.
        """
        lb_idx, ub_idx = max(self._drag_idx - 1, 0), min(self._drag_idx + 1, self._x_obj.shape[0] - 1)

        while lb_idx >= 0 and (self._x_obj.mask[lb_idx] or self._y_obj.mask[lb_idx]):
            lb_idx -= 1

        while ub_idx < self._x_obj.shape[0] and (self._x_obj.mask[ub_idx] or self._y_obj.mask[ub_idx]):
            ub_idx += 1

        if lb_idx != -1:
            lb_x, ub_x = min(drag_x, self._x_obj[lb_idx]), max(drag_x, self._x_obj[lb_idx])
            lb_y, ub_y = min(drag_y, self._y_obj[lb_idx]), max(drag_y, self._y_obj[lb_idx])
        else:
            lb_x = ub_x = drag_x
            lb_y = ub_y = drag_y

        if ub_idx != self._x_obj.shape[0]:
            lb_x, ub_x = min(lb_x, self._x_obj[ub_idx]), max(ub_x, self._x_obj[ub_idx])
            lb_y, ub_y = min(lb_y, self._y_obj[ub_idx]), max(ub_y, self._y_obj[ub_idx])

        qp = QPainter()
        qp.begin(self._background)

        self._restoreBitmap(qp)
        self._saveBitmap(lb_x, lb_y, ub_x, ub_y)

        pen = QPen(self._line_color, 1, Qt.SolidLine)
        qp.setPen(pen)
        if lb_idx != -1:
            x1, y1 = self._x_obj[lb_idx], self._y_obj[lb_idx]
            x2, y2 = drag_x, drag_y
            qp.drawLine(x1, y1, x2, y2)
        if ub_idx != self._x_obj.shape[0]:
            x1, y1 = drag_x, drag_y
            x2, y2 = self._x_obj[ub_idx], self._y_obj[ub_idx]
            qp.drawLine(x1, y1, x2, y2)

        qp.end()

    def _dragPoint(self, drag_x, drag_y):
        """
        Drag a point [private method].
        drag_x: x coordinate in pixels of the drag point
        drag_y: y coordinate in pixels of the drag point
        """
        qp = QPainter()
        qp.begin(self._background)

        self._restoreBitmap(qp)
        self._saveBitmap(drag_x, drag_y, drag_x, drag_y)

        pen = QPen(self._line_color, 1, Qt.SolidLine)
        qp.setPen(pen)

        qp.drawEllipse(QPointF(drag_x, drag_y), 3, 3)
        qp.end()
       
    def release(self, rls_x, rls_y, restrictions=None):
        """
        Release the drag, which finishes the drag operation.
        rls_x:  x coordinate in pixels of the release point
        rls_y:  y coordinate in pixels of the release point
        restrictions: [See the release argument to the drag method]
        """
        if self._drag_idx is None:
            return

        start_x, start_y = self._click_start
        if rls_x == start_x and rls_y == start_y:
            rls_x = self._x_obj[self._drag_idx]
            rls_y = self._y_obj[self._drag_idx]
        else:
            if self._lock_dim == 'x':
                rls_x = self._x_obj[self._drag_idx]
            elif self._lock_dim == 'y':
                rls_y = self._y_obj[self._drag_idx]

            if restrictions:
                rls_x, rls_y = restrictions(rls_x, rls_y)

            self._x_obj[self._drag_idx] = rls_x
            self._y_obj[self._drag_idx] = rls_y

        drag_idx = self._drag_idx
        self._drag_idx = None
        self._save_bitmap = None
        self._click_start = None
        return drag_idx, rls_x, rls_y

    def setBackground(self, background):
        """
        Change the background pixmap.
        background: A QPixmap containing the new background
        """
        self._background = background

    def setCoords(self, x_obj, y_obj):
        """
        Change the coordinates.
        x_obj:  x coordinate(s) in pixels of the new object
        y_obj:  y coordinate(s) in pixesl of the new object
        """
        self._x_obj = x_obj
        self._y_obj = y_obj

        if type(self._x_obj) != np.ma.MaskedArray:
            self._x_obj = np.ma.array([ self._x_obj ])
        if type(self._y_obj) != np.ma.MaskedArray:
            self._y_obj = np.ma.array([ self._y_obj ])

    def getBackground(self):
        """
        Returns the current background as a QPixmap
        """
        return self._background

    def getCoords(self):
        """
        Returns the current coordinates as numpy arrays
        """
        return self._x_obj, self._y_obj

    def isDragging(self):
        """
        Returns a boolean specifying whether or not this object is in the 
        middle of a drag operation.
        """
        return self._drag_idx is not None

    def _saveBitmap(self, lb_x, lb_y, ub_x, ub_y):
        """
        Save a section of the background image [private method]. A 5 px buffer is applied 
            in all directions.
        lb_x:   Lower bound on the x coordinate in pixels.
        ub_x:   Upper bound on the x coordinate in pixels.
        lb_y:   Lower bound on the y coordinate in pixels.
        ub_y:   Upper bound on the y coordinate in pixels.
        """
        origin = QPoint(max(lb_x - self._drag_buffer, 0), max(lb_y - self._drag_buffer, 0))
        size = QSize(ub_x - lb_x + 2 * self._drag_buffer, ub_y - lb_y + 2 * self._drag_buffer)
        bmap = self._background.copy(QRect(origin, size))
        self._save_bitmap = (origin, size, bmap)

    def _restoreBitmap(self, qp):
        """
        Restore the section of the background image saved by _saveBitmap [private method].
        qp: A QPainter to use to draw the bitmap.
        """
        if self._save_bitmap is not None:
            (origin, size, bmap) = self._save_bitmap
            qp.drawPixmap(origin, bmap, QRect(QPoint(0, 0), size))





from qtpy.QtGui import *
from qtpy.QtCore import *

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser
import os

class Config(QObject):
    """
    The Config class keeps track of the configuration in a QObject, so that it 
    can be passed via signals and slots.
    """
    def __init__(self, file_name):
        """
        Construct a Config object.
        file_name:  The name of the file to store configuration information in.
        """
        super(Config, self).__init__()

        self._file_name = file_name
        self._cfg = ConfigParser.RawConfigParser()
        if os.path.exists(file_name):
            self._cfg.read(file_name)

    def __getitem__(self, item):
        """
        Get an item from the configuration. The call should look like this:
        value = config['section', 'field']
        """
        section, field = item
        return self._cfg.get(section, field)

    def __setitem__(self, item, value):
        """
        Set an item in the configuration, overwriting what's already there. The call should look like this:
        config['section', 'field'] = value
        """
        section, field = item
        if not self._cfg.has_section(section):
            self._cfg.add_section(section)

        self._cfg.set(section, field, value)

    def __contains__(self, item):
        """
        Check to see if an item is in the configuration. The call should look like this:
        ('field', 'section') in config
        """
        section, field = item
        return self._cfg.has_option(section, field)

    def __iter__(self):
        """
        Iterate over all the fields in the configuration. The call might look like this:
        for (section, field), value in config: # ...
        """
        for section in self._cfg.sections():
            for field in self._cfg.options(section):
                yield (section, field), self[section, field]

    def initialize(self, cfg_dict):
        """
        Add items in the configuration. This function doesn't modify items that already exist.
        cfg_dict:   A dictionary of configuration items. Keys are (section, field) tuples, and the values
                    are the configuration values.
        """
        for item, value in cfg_dict.items():
            if item not in self:
                self[item] = value

    def toFile(self):
        """
        Write the configuration to the file.
        """
        self._cfg.write(open(self._file_name, 'w'))



from . import thermo, utils, interp, params, constants
import numpy as np
import logging

## Routines implemented in Python by Greg Blumberg - CIMMS and Kelton Halbert (OU SoM)
## wblumberg@ou.edu, greg.blumberg@noaa.gov, kelton.halbert@noaa.gov, keltonhalbert@ou.edu

def heat_index(temp, rh):
    '''
        Heat Index Equation

        Computes the heat index using the equation obtained by performing multiple linear
        regression on the table in Steadman 1979.
    
        Referenced from: http://www.srh.noaa.gov/images/ffc/pdf/ta_htindx.PDF

        Parameters
        ----------
        temp : number
            temperature (F)
        rh : number
            relative humidity (%)

        Returns
        -------
        heat_index : number
            heat index value in (F)
    '''
    if temp < 40:
        return temp

    hi = 0.5 * ( temp + 61.0 + ((temp - 68.0) * 1.2) + (rh * 0.094))
    avg = (hi + temp)/2.
    if avg < 80:
        return hi
    #temp = thermo.ctof(prof.tmpc[prof.get_sfc()])
    #rh = thermo.relh(prof.pres[prof.get_sfc()], temp, prof.dwpc[prof.get_sfc()])
    heat_index = -42.379 + (2.04901523 * temp) + (10.14333127 * rh) - (0.22475541 * temp * rh) - (6.83783e-3 * np.power(temp,2)) \
                 - (5.481717e-2 * np.power(rh, 2)) + (1.22874e-3 * rh * np.power(temp,2)) + (8.5282e-4 * temp * np.power(rh, 2)) \
                 - (1.99e-6 * np.power(rh, 2) * np.power(temp, 2))

    if rh < 13 and temp > 80 and temp < 112:
        adjustment = ((13-rh)/4.) * np.sqrt((17 - np.abs(temp - 95))/17.)
        heat_index = heat_index - adjustment
    elif rh > 85 and temp > 80 and temp < 87:
        adjustment = ((rh - 85)/10.) * ((87 - temp)/5.)
        heat_index = heat_index + adjustment  
    return heat_index

def wind_chill(prof):
    '''
        Surface Wind Chill Equation

        Computes wind chill at the surface data point in the profile object
        using the equation found at:

        www.nws.noaa.gov/os/windchill/index.shtml

        Parameters
        ----------
        prof : profile object
            Profile object

        Returns
        -------
        wind_chill : number
            wind chill value in (F)
    '''
    # Needs to be tested

    sfc_temp = thermo.ctof(prof.tmpc[prof.get_sfc()])
    sfc_wspd = utils.KTS2MPH(prof.wspd[prof.get_sfc()])

    wind_chill = 35.74 + (0.6215*sfc_temp) - (35.75*(sfc_wspd**0.16)) + \
                 0.4275 * (sfc_temp) * (sfc_wspd**0.16)
    return wind_chill

def init_phase(prof):
    '''
        Inital Precipitation Phase
        Adapted from SHARP code donated by Rich Thompson (SPC)

        This function determines the initial phase of any precipitation source in the profile.
        It does this either by finding a source of precipitation by searching for the highest 50 mb 
        layer that has a relative humidity greater than 80 percent at the top and the bottom
        of the layer.  This layer may be found either in the lowest 5 km of the profile, and if
        an OMEG profile is specified in the profile object, it will search for the layers with
        upward motion.

        The precipitation type is determined by using a.) the interpolated temperature in the middle
        of the precipitation source layer and b.) set temperature thresholds to determine the 
        precipitation type.  The type may be "Rain", "Freezing Rain", "ZR/S Mix", or "Snow".

        Parameters
        ----------
        prof : profile object
            Profile object (omega profile optional)

        Returns
        -------
        plevel : number
            the pressure level of the precipitation source (mb)
        phase : int
            the phase type of the precipitation (int), phase = 0 for "Rain", phase = 1 for "Freezing Rain" or "ZR/S Mix", phase = 3 for "Snow"
        tmp : number
            the temperature at the level that is the precipitation source (C)
        st : str
            a string naming the precipitation type

    '''
    # Needs to be tested

    plevel = 0
    phase = -1

    # First, determine whether Upward VVELS are available.  If they are,  
    # use them to determine level where precipitation will develop.
    avail = np.ma.where(prof.omeg < .1)[0]

    hght_agl = interp.to_agl(prof, prof.hght)
    if len(avail) < 5:
        # No VVELS...must look for saturated level 
        # Find the highest near-saturated 50mb layer below 5km agl
        below_5km_idx = np.ma.where((hght_agl < 5000.) &\
                                    (hght_agl >= 0))[0]

    else:
        # Use the VV to find the source of precip.
        below_5km_idx = np.ma.where((hght_agl < 5000.) &\
                                    (hght_agl >= 0) &\
                                    (prof.omeg <= 0))[0]

    # Compute the RH at the top and bottom of 50 mb layers
    rh = thermo.relh(prof.pres, prof.tmpc, prof.dwpc)[below_5km_idx]
    sats = np.ma.where(rh > 80)[0]
    new_pres = prof.pres[below_5km_idx][sats] + 50.
    new_temp = interp.temp(prof, new_pres)
    new_dwpt = interp.dwpt(prof, new_pres)
    rh_plus50 = thermo.relh(new_pres, new_temp, new_dwpt)
    # Find layers where the RH is >80% at the top and bottom
    layers_idx = np.ma.where(rh_plus50 > 80)[0]

    if len(layers_idx) == 0:
        # Found no precipitation source layers
        st = "N/A"
        return prof.missing, phase, prof.missing, st

    # Find the highest layer up via the largest index
    top_most_layer = np.ma.max(layers_idx)
    plevel = new_pres[top_most_layer] - 25.

    # Determine the initial precip type based on the temp in the layer
    tmp = interp.temp(prof, plevel)
    if tmp > 0:
        phase = 0
        st = "Rain"
    elif tmp <= 0 and tmp > -5:
        phase = 1
        st = "Freezing Rain"
    elif tmp <=-5 and tmp > -9:
        phase = 1
        st = "ZR/S Mix"
    elif tmp <= -9:
        phase = 3
        st = "Snow"
    else:
        st = "N/A"

    return plevel, phase, tmp, st

def posneg_temperature(prof, start=-1):
    '''
        Positive/Negative Temperature profile
        Adapted from SHARP code donated by Rich Thompson (SPC)

        Description:
        This routine calculates the positive (above 0 C) and negative (below 0 C)
        areas of the temperature profile starting from a specified pressure (start).
        If the specified pressure is not given, this routine calls init_phase()
        to obtain the pressure level the precipitation expected to fall begins at.

        This is an routine considers only the temperature profile as opposed to the wet-bulb
        profile.

        Parameters
        ----------
        prof : profile object
            Profile object
        start : number
            the pressure level the precipitation originates from (found by calling init_phase()) (mb)

        Returns
        -------
        pos : float
            the positive area (> 0 C) of the wet-bulb profile (J/kg)
        neg : float
            the negative area (< 0 C) of the wet-bulb profile (J/kg)
        top : float
            the top of the precipitation layer pressure (mb)
        bot : float
            the bottom of the precipitation layer pressure (mb)

    '''
    # Needs to be tested
    
    # If there is no sounding, don't compute anything
    if utils.QC(interp.temp(prof, 500)) == False and utils.QC(interp.temp(prof, 850)) == False:
        return np.ma.masked, np.ma.masked, np.ma.masked, np.ma.masked

    # Find lowest obs in layer
    lower = prof.pres[prof.get_sfc()]
    lptr  = prof.get_sfc()

    # Find the highest obs in the layer
    if start == -1:
        lvl, phase, st = init_phase(prof)
        if lvl > 0:
            upper = lvl
        else:
            upper = 500.
    else:
        upper = start

    # Find the level where the pressure is just greater than the upper pressure
    idxs = np.where(prof.pres > upper)[0]
    if len(idxs) == 0:
        uptr = 0
    else:
        uptr = idxs[-1]

    # Start with the top layer
    pe1 = upper;
    h1 =  interp.hght(prof, pe1)
    te1 = interp.temp(prof, pe1)
    tp1 = 0

    warmlayer = coldlayer = lyre = totp = totn = tote = ptop = pbot = lyrlast = 0

    for i in np.arange(uptr, lptr-1, -1):
        pe2 = prof.pres[i]
        h2 = prof.hght[i]
        te2 = interp.temp(prof, pe2)
        tp2 = 0
        tdef1 = (0 - te1) / thermo.ctok(te1);
        tdef2 = (0 - te2) / thermo.ctok(te2);
        lyrlast = lyre;
        lyre = 9.8 * (tdef1 + tdef2) / 2.0 * (h2 - h1);

        # Has a warm layer been found yet?
        if te2 > 0:
            if warmlayer == 0:
                warmlayer = 1
                ptop = pe2

        # Has a cold layer been found yet?
        if te2 < 0:
            if warmlayer == 1 and coldlayer == 0:
                coldlayer = 1
                pbot = pe2

        if warmlayer > 0:
            if lyre > 0:
                totp += lyre
            else:
                totn += lyre
            tote += lyre

        pelast = pe1
        pe1 = pe2
        h1 = h2
        te1 = te2
        tp1 = tp2
    
    if warmlayer == 1 and coldlayer == 1:
        pos = totp
        neg = totn
        top = ptop
        bot = pbot
    else:
        neg = 0
        pos = 0
        bot = 0
        top = 0

    return pos, neg, top, bot


def posneg_wetbulb(prof, start=-1):
    '''
        Positive/Negative Wetbulb profile
        Adapted from SHARP code donated by Rich Thompson (SPC)

        This routine calculates the positive (above 0 C) and negative (below 0 C)
        areas of the wet bulb profile starting from a specified pressure (start).
        If the specified pressure is not given, this routine calls init_phase()
        to obtain the pressure level the precipitation expected to fall begins at.

        This is an routine considers the wet-bulb profile instead of the temperature profile
        in case the profile beneath the profile beneath the falling precipitation becomes saturated.

        Parameters
        ----------
        prof : profile object
            Profile object
        start : number
            the pressure level the precipitation originates from (found by calling init_phase()) (mb)

        Returns
        -------
        pos : float
            the positive area (> 0 C) of the wet-bulb profile (J/kg)
        neg : float
            the negative area (< 0 C) of the wet-bulb profile (J/kg)
        top : float
            the top of the precipitation layer pressure (mb)
        bot : float
            the bottom of the precipitation layer pressure (mb)

    '''
    # Needs to be tested

    # If there is no sounding, don't compute anything
    if utils.QC(interp.temp(prof, 500)) == False and utils.QC(interp.temp(prof, 850)) == False:
        return np.ma.masked, np.ma.masked, np.ma.masked, np.ma.masked

    # Find lowest obs in layer
    lower = prof.pres[prof.get_sfc()]
    lptr  = prof.get_sfc()

    # Find the highest obs in the layer
    if start == -1:
        lvl, phase, st = init_phase(prof)
        if lvl > 0:
            upper = lvl
        else:
            upper = 500.
    else:
        upper = start

    # Find the level where the pressure is just greater than the upper pressure
    idxs = np.where(prof.pres > upper)[0]
    if len(idxs) == 0:
        uptr = 0
    else:
        uptr = idxs[-1]

    # Start with the upper layer
    pe1 = upper;
    h1 =  interp.hght(prof, pe1);
    te1 = thermo.wetbulb(pe1, interp.temp(prof, pe1), interp.dwpt(prof, pe1))
    tp1 = 0

    warmlayer = coldlayer = lyre = totp = totn = tote = ptop = pbot = lyrlast = 0

    for i in np.arange(uptr, lptr-1, -1):
        pe2 = prof.pres[i]
        h2 = prof.hght[i]
        te2 = thermo.wetbulb(pe2, interp.temp(prof, pe2), interp.dwpt(prof, pe2))
        tp2 = 0
        tdef1 = (0 - te1) / thermo.ctok(te1);
        tdef2 = (0 - te2) / thermo.ctok(te2);
        lyrlast = lyre;
        lyre = 9.8 * (tdef1 + tdef2) / 2.0 * (h2 - h1);

        # Has a warm layer been found yet?
        if te2 > 0:
            if warmlayer == 0:
                warmlayer = 1
                ptop = pe2

        # Has a cold layer been found yet?
        if te2 < 0:
            if warmlayer == 1 and coldlayer == 0:
                coldlayer = 1
                pbot = pe2

        if warmlayer > 0:
            if lyre > 0:
                totp += lyre
            else:
                totn += lyre
            tote += lyre

        pelast = pe1
        pe1 = pe2
        h1 = h2
        te1 = te2
        tp1 = tp2
    
    if warmlayer == 1 and coldlayer == 1:
        pos = totp
        neg = totn
        top = ptop
        bot = pbot
    else:
        neg = 0
        pos = 0
        bot = 0
        top = 0

    return pos, neg, top, bot

def best_guess_precip(prof, init_phase, init_lvl, init_temp, tpos, tneg):
    '''
        Best Guess Precipitation type
        Adapted from SHARP code donated by Rich Thompson (SPC)

        This algorithm utilizes the output from the init_phase() and posneg_temperature()
        functions to make a best guess at the preciptation type one would observe
        at the surface given a thermodynamic profile.

        Precipitation Types Supported:
        * None
        * Rain
        * Snow
        * Sleet and Snow
        * Sleet
        * Freezing Rain/Drizzle
        * Unknown

        Parameters
        ----------
        prof : profile object
            Profile object
        init_phase : int
            the initial phase of the precipitation (see 2nd value returned from init_phase())
        init_lvl : float
            the initial level of the precipitation source (mb) (see 1st value returned from init_phase())
        init_temp : float
            the initial level of the precipitation source (C) (see 3rd value returned from init_phase())
        tpos : float
            the positive area (> 0 C) in the temperature profile (J/kg)

        Returns
        -------
        precip_type : str
            the best guess precipitation type
    '''
    # Needs to be tested

    precip_type = None

    # Case: No precip
    if init_phase < 0:
        precip_type = "None."

    # Case: Always too warm - Rain
    elif init_phase == 0 and tneg >=0 and prof.tmpc[prof.get_sfc()] > 0:
        precip_type = "Rain."

    # Case: always too cold
    elif init_phase == 3 and tpos <= 0 and prof.tmpc[prof.get_sfc()] <= 0:
        precip_type = "Snow."

    # Case: ZR too warm at sfc - Rain
    elif init_phase == 1 and tpos <= 0 and prof.tmpc[prof.get_sfc()] > 0:
        precip_type = "Rain."

    # Case: non-snow init...always too cold - Initphase & sleet
    elif init_phase == 1 and tpos <= 0 and prof.tmpc[prof.get_sfc()] <= 0:
        #print interp.to_agl(prof, interp.hght(prof, init_lvl))
        if interp.to_agl(prof, interp.hght(prof, init_lvl)) >= 3000:
            if init_temp <= -4:
                precip_type = "Sleet and Snow."
            else:
                precip_type = "Sleet."
        else:
            precip_type = "Freezing Rain/Drizzle."

    # Case: Snow...but warm at sfc
    elif init_phase == 3 and tpos <= 0 and prof.tmpc[prof.get_sfc()] > 0:
        if prof.tmpc[prof.get_sfc()] > 4:
            precip_type = "Rain."
        else:
            precip_type = "Snow."
   
    # Case: Warm layer.
    elif tpos > 0:
        x1 = tpos
        y1 = -tneg
        y2 = (0.62 * x1) + 60.0
        if y1 > y2:
            precip_type = "Sleet."
        else:
            if prof.tmpc[prof.get_sfc()] <= 0:
                precip_type = "Freezing Rain."
            else:
                precip_type = "Rain."
    else:
        precip_type = "Unknown."

    return precip_type

def possible_watch(prof, use_left=False):
    '''
        Possible Weather/Hazard/Watch Type
        
        This function generates a list of possible significant weather types
        one can expect given a Profile object. (Currently works only for ConvectiveProfile.)

        These possible weather types are computed via fuzzy logic through set thresholds that
        have been found through a.) analyzing ingredients within the profile and b.) combining those ingredients
        with forecasting experience to produce a suggestion of what hazards may exist.  Some of the logic is 
        based on experience, some of it is based on actual National Weather Service criteria.

        This function has not been formally verified and is not meant to be comprehensive nor
        a source of strict guidance for weather forecasters.  As always, the raw data is to be 
        consulted.

        Wx Categories (ranked in terms of severity):
        * PDS TOR
        * TOR
        * MRGL TOR
        * SVR
        * MRGL SVR
        * FLASH FLOOD
        * BLIZZARD
        * EXCESSIVE HEAT
    
        Suggestions for severe/tornado thresholds were contributed by Rich Thompson - NOAA Storm Prediction Center

        Parameters
        ----------
        prof : profile object
            ConvectiveProfile object
        use_left : bool
            If True, uses the parameters computed from the left-mover bunkers vector to decide the watch type. If False,
            uses parameters from the right-mover vector. The default is False.

        Returns
        -------
        watch_types : numpy array
            strings containing the weather types in code
    '''
        
    watch_types = []
    
    lr1 = params.lapse_rate( prof, 0, 1000, pres=False )
    if use_left:
        stp_eff = prof.left_stp_cin
        stp_fixed = prof.left_stp_fixed
        srw_4_6km = utils.mag(prof.left_srw_4_6km[0],prof.left_srw_4_6km[1])
        esrh = prof.left_esrh[0]
        srh1km = prof.left_srh1km[0]
    else:
        stp_eff = prof.right_stp_cin
        stp_fixed = prof.right_stp_fixed
        srw_4_6km = utils.mag(prof.right_srw_4_6km[0],prof.right_srw_4_6km[1])
        esrh = prof.right_esrh[0]
        srh1km = prof.right_srh1km[0]

    if prof.latitude < 0:
        stp_eff = -stp_eff
        stp_fixed = -stp_fixed
        esrh = -esrh
        srh1km = -srh1km

    sfc_8km_shear = utils.mag(prof.sfc_8km_shear[0],prof.sfc_8km_shear[1])

    if stp_eff >= 3 and stp_fixed >= 3 and srh1km >= 200 and esrh >= 200 and srw_4_6km >= 15.0 and \
        sfc_8km_shear > 45.0 and prof.sfcpcl.lclhght < 1000. and prof.mlpcl.lclhght < 1200 and lr1 >= 5.0 and \
        prof.mlpcl.bminus > -50 and prof.ebotm == 0:
        watch_types.append("PDS TOR")
    elif (stp_eff >= 3 or stp_fixed >= 4) and prof.mlpcl.bminus > -125. and prof.ebotm == 0:
        watch_types.append("TOR")
    elif (stp_eff >= 1 or stp_fixed >= 1) and (srw_4_6km >= 15.0 or sfc_8km_shear >= 40) and \
        prof.mlpcl.bminus > -50 and prof.ebotm == 0:
        watch_types.append("TOR")
    elif (stp_eff >= 1 or stp_fixed >= 1) and ((prof.low_rh + prof.mid_rh)/2. >= 60) and lr1 >= 5.0 and \
        prof.mlpcl.bminus > -50 and prof.ebotm == 0:
        watch_types.append("TOR")
    elif (stp_eff >= 1 or stp_fixed >= 1) and prof.mlpcl.bminus > -150 and prof.ebotm == 0.:
        watch_types.append("MRGL TOR")
    elif (stp_eff >= 0.5 and esrh >= 150) or (stp_fixed >= 0.5 and srh1km >= 150) and \
        prof.mlpcl.bminus > -50 and prof.ebotm == 0.:
        watch_types.append("MRGL TOR")

    #SVR LOGIC
    if use_left:
        scp = prof.left_scp
    else:
        scp = prof.right_scp

    if (stp_fixed >= 1.0 or scp >= 4.0 or stp_eff >= 1.0) and prof.mupcl.bminus >= -50:
        watch_types.append("SVR")
    elif scp >= 2.0 and (prof.ship >= 1.0 or prof.dcape >= 750) and prof.mupcl.bminus >= -50:
        watch_types.append("SVR")
    elif prof.sig_severe >= 30000 and prof.mmp >= 0.6 and prof.mupcl.bminus >= -50:
        watch_types.append("SVR")
    elif prof.mupcl.bminus >= -75.0 and (prof.wndg >= 0.5 or prof.ship >= 0.5 or scp >= 0.5):
        watch_types.append("MRGL SVR")
    
    # Flash Flood Watch PWV is larger than normal and cloud layer mean wind speeds are slow
    # This is trying to capture the ingredients of moisture and advection speed, but cannot
    # handle precipitation efficiency or vertical motion

    # Likely is good for handling slow moving MCSes.
    pw_climo_flag = prof.pwv_flag
    pwat = prof.pwat
    upshear = utils.comp2vec(prof.upshear_downshear[0],prof.upshear_downshear[1])
    if pw_climo_flag >= 2 and upshear[1] < 25:
        watch_types.append("FLASH FLOOD")
    #elif pwat > 1.3 and upshear[1] < 25:
    #    watch_types.append("FLASH FLOOD")
    
    # Blizzard if sfc winds > 35 mph and precip type detects snow 
    # Still needs to be tied into the 
    sfc_wspd = utils.KTS2MPH(prof.wspd[prof.get_sfc()])
    if sfc_wspd > 35. and prof.tmpc[prof.get_sfc()] <= 0 and "Snow" in prof.precip_type:
        watch_types.append("BLIZZARD")
    
    # Wind Chill (if wind chill gets below -20 F)
    # TODO: Be reinstated in future releases if the logic becomes a little more solid.
    #if wind_chill(prof) < -20.:
    #    watch_types.append("WIND CHILL")
    
    # Fire WX (sfc RH < 30% and sfc_wind speed > 15 mph) (needs to be updated to include SPC Fire Wx Indices)
    # TODO: Be reinstated in future releases once the logic becomes a little more solid
    #if sfc_wspd > 15. and thermo.relh(prof.pres[prof.get_sfc()], prof.tmpc[prof.get_sfc()], prof.dwpc[prof.get_sfc()]) < 30. :
        #watch_types.append("FIRE WEATHER")
    
    # Excessive Heat (use the heat index calculation (and the max temperature algorithm))
    temp = thermo.ctof(prof.tmpc[prof.get_sfc()])
    rh = thermo.relh(prof.pres[prof.get_sfc()], temp, prof.dwpc[prof.get_sfc()])
    hi = heat_index(temp, rh)
    if hi > 105.:
        watch_types.append("EXCESSIVE HEAT")
    
    # Freeze (checks to see if wetbulb is below freezing and temperature isn't and wind speeds are low)
    # Still in testing.  To be reinstated in future releases.
    #if thermo.ctof(prof.dwpc[prof.get_sfc()]) <= 32. and thermo.ctof(prof.wetbulb[prof.get_sfc()]) <= 32 and prof.wspd[prof.get_sfc()] < 5.:
    #    watch_types.append("FREEZE")
    
    watch_types.append("NONE")
    
    return np.asarray(watch_types)




# Copyright (c) 2014,2015,2016,2017 MetPy Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Make Skew-T Log-P based plots.

Contain tools for making Skew-T Log-P plots, including the base plotting class,
`SkewT`, as well as a class for making a `Hodograph`.
"""

import matplotlib
from matplotlib.axes import Axes
import matplotlib.axis as maxis
from matplotlib.collections import LineCollection
import matplotlib.colors as mcolors
from matplotlib.patches import Circle
from matplotlib.projections import register_projection
import matplotlib.spines as mspines
from matplotlib.ticker import MultipleLocator, NullFormatter, ScalarFormatter
from matplotlib.pyplot import title
import matplotlib.transforms as transforms
import numpy as np
from . import sharptab as tab

class SkewXTick(maxis.XTick):
    r"""Make x-axis ticks for Skew-T plots.

    This class adds to the standard :class:`matplotlib.axis.XTick` dynamic checking
    for whether a top or bottom tick is actually within the data limits at that part
    and draw as appropriate. It also performs similar checking for gridlines.
    """

    def update_position(self, loc):
        """Set the location of tick in data coords with scalar *loc*."""
        # This ensures that the new value of the location is set before
        # any other updates take place.
        self._loc = loc
        super(SkewXTick, self).update_position(loc)

    def _has_default_loc(self):
        return self.get_loc() is None

    def _need_lower(self):
        return (self._has_default_loc()
                or transforms.interval_contains(self.axes.lower_xlim, self.get_loc()))

    def _need_upper(self):
        return (self._has_default_loc()
                or transforms.interval_contains(self.axes.upper_xlim, self.get_loc()))

    @property
    def gridOn(self):  # noqa: N802
        """Control whether the gridline is drawn for this tick."""
        return (self._gridOn and (self._has_default_loc()
                or transforms.interval_contains(self.get_view_interval(), self.get_loc())))

    @gridOn.setter
    def gridOn(self, value):  # noqa: N802
        self._gridOn = value

    @property
    def tick1On(self):  # noqa: N802
        """Control whether the lower tick mark is drawn for this tick."""
        return self._tick1On and self._need_lower()

    @tick1On.setter
    def tick1On(self, value):  # noqa: N802
        self._tick1On = value

    @property
    def label1On(self):  # noqa: N802
        """Control whether the lower tick label is drawn for this tick."""
        return self._label1On and self._need_lower()

    @label1On.setter
    def label1On(self, value):  # noqa: N802
        self._label1On = value

    @property
    def tick2On(self):  # noqa: N802
        """Control whether the upper tick mark is drawn for this tick."""
        return self._tick2On and self._need_upper()

    @tick2On.setter
    def tick2On(self, value):  # noqa: N802
        self._tick2On = value

    @property
    def label2On(self):  # noqa: N802
        """Control whether the upper tick label is drawn for this tick."""
        return self._label2On and self._need_upper()

    @label2On.setter
    def label2On(self, value):  # noqa: N802
        self._label2On = value

    def get_view_interval(self):
        """Get the view interval."""
        return self.axes.xaxis.get_view_interval()


class SkewXAxis(maxis.XAxis):
    r"""Make an x-axis that works properly for Skew-T plots.

    This class exists to force the use of our custom :class:`SkewXTick` as well
    as provide a custom value for interview that combines the extents of the
    upper and lower x-limits from the axes.
    """

    def _get_tick(self, major):
        return SkewXTick(self.axes, None, '', major=major)

    def get_view_interval(self):
        """Get the view interval."""
        return self.axes.upper_xlim[0], self.axes.lower_xlim[1]


class SkewSpine(mspines.Spine):
    r"""Make an x-axis spine that works properly for Skew-T plots.

    This class exists to use the separate x-limits from the axes to properly
    locate the spine.
    """

    def _adjust_location(self):
        pts = self._path.vertices
        if self.spine_type == 'top':
            pts[:, 0] = self.axes.upper_xlim
        else:
            pts[:, 0] = self.axes.lower_xlim


class SkewXAxes(Axes):
    r"""Make a set of axes for Skew-T plots.

    This class handles registration of the skew-xaxes as a projection as well as setting up
    the appropriate transformations. It also makes sure we use our instances for spines
    and x-axis: :class:`SkewSpine` and :class:`SkewXAxis`. It provides properties to
    facilitate finding the x-limits for the bottom and top of the plot as well.
    """

    # The projection must specify a name.  This will be used be the
    # user to select the projection, i.e. ``subplot(111,
    # projection='skewx')``.
    name = 'skewx'

    def __init__(self, *args, **kwargs):
        r"""Initialize `SkewXAxes`.

        Parameters
        ----------
        args : Arbitrary positional arguments
            Passed to :class:`matplotlib.axes.Axes`

        position: int, optional
            The rotation of the x-axis against the y-axis, in degrees.

        kwargs : Arbitrary keyword arguments
            Passed to :class:`matplotlib.axes.Axes`

        """
        # This needs to be popped and set before moving on
        self.rot = kwargs.pop('rotation', 30)
        Axes.__init__(self, *args, **kwargs)

    def _init_axis(self):
        # Taken from Axes and modified to use our modified X-axis
        self.xaxis = SkewXAxis(self)
        self.spines['top'].register_axis(self.xaxis)
        self.spines['bottom'].register_axis(self.xaxis)
        self.yaxis = maxis.YAxis(self)
        self.spines['left'].register_axis(self.yaxis)
        self.spines['right'].register_axis(self.yaxis)

    def _gen_axes_spines(self, locations=None, offset=0.0, units='inches'):
        # pylint: disable=unused-argument
        spines = {'top': SkewSpine.linear_spine(self, 'top'),
                  'bottom': mspines.Spine.linear_spine(self, 'bottom'),
                  'left': mspines.Spine.linear_spine(self, 'left'),
                  'right': mspines.Spine.linear_spine(self, 'right')}
        return spines

    def _set_lim_and_transforms(self):
        """Set limits and transforms.

        This is called once when the plot is created to set up all the
        transforms for the data, text and grids.

        """
        # Get the standard transform setup from the Axes base class
        Axes._set_lim_and_transforms(self)

        # Need to put the skew in the middle, after the scale and limits,
        # but before the transAxes. This way, the skew is done in Axes
        # coordinates thus performing the transform around the proper origin
        # We keep the pre-transAxes transform around for other users, like the
        # spines for finding bounds
        self.transDataToAxes = (self.transScale
                                + (self.transLimits
                                   + transforms.Affine2D().skew_deg(self.rot, 0)))

        # Create the full transform from Data to Pixels
        self.transData = self.transDataToAxes + self.transAxes

        # Blended transforms like this need to have the skewing applied using
        # both axes, in axes coords like before.
        self._xaxis_transform = (
            transforms.blended_transform_factory(self.transScale + self.transLimits,
                                                 transforms.IdentityTransform())
            + transforms.Affine2D().skew_deg(self.rot, 0)) + self.transAxes

    @property
    def lower_xlim(self):
        """Get the data limits for the x-axis along the bottom of the axes."""
        return self.axes.viewLim.intervalx

    @property
    def upper_xlim(self):
        """Get the data limits for the x-axis along the top of the axes."""
        return self.transDataToAxes.inverted().transform([[0., 1.], [1., 1.]])[:, 0]


# Now register the projection with matplotlib so the user can select
# it.
register_projection(SkewXAxes)
pb_plot=1050
pt_plot=100
dp_plot=10
plevs_plot = presvals = np.arange(pb_plot,pt_plot-1,-dp_plot)


def draw_dry_adiabats(ax, tmin=-50, tmax=210, delta=10, color='r', alpha=.2):
    # plot the dry adiabats
    for t in np.arange(tmin,tmax,delta):
        ax.semilogy(thetas(t, presvals), presvals, '-', color=color, alpha=alpha)
    return ax

def draw_mixing_ratio_lines(ax, spacing=[2,4,10,12,14,16,18,20], color='g', lw=.7):
    # plot the mixing ratio lines
    for w in spacing:
        line = tab.thermo.temp_at_mixrat(w, np.arange(1000,600,-1))
        ax.semilogy(line, np.arange(1000,600,-1), '--', color=color, lw=lw)
    return ax

def draw_moist_adiabats(ax, tmin=-50, tmax=50, delta=10, color='k', alpha=0.5):
    # plot the moist adiabats
    for i in np.arange(tmin,tmax,delta):
        t = []
        for pres in np.arange(1000,90,-10):
            t.append(tab.thermo.wetlift(1000,i,pres))
        ax.semilogy(t, np.arange(1000,90,-10), color=color, lw=1, alpha=alpha)
    return ax

def draw_title(ax, t):
    title(t, fontsize=14, loc='left')
    return ax

# Routine to plot the axes for the wind profile
def plot_wind_axes(axes):
    # plot wind barbs
    draw_wind_line(axes)
    axes.set_axis_off()
    axes.axis([-1,1,pb_plot,pt_plot])

     
# Routine to plot the wind barbs.
def plot_wind_barbs(axes, p, u, v):
    for i in np.arange(0,len(p)):
        if (p[i] > pt_plot):
            if np.ma.is_masked(v[i]) is True:
                continue
            axes.barbs(0,p[i],u[i],v[i], length=7, clip_on=False, linewidth=1)

# Routine to draw the line for the wind barbs
def draw_wind_line(axes):
    wind_line = []
    for p in plevs_plot:
        wind_line.append(0)
    axes.semilogy(wind_line, plevs_plot, color='black', linewidth=.5)

# Routine to calculate the dry adiabats.
def thetas(theta, presvals):
    return ((theta + tab.thermo.ZEROCNK) / (np.power((1000. / presvals),tab.thermo.ROCP))) - tab.thermo.ZEROCNK

def plot_sig_levels(ax, prof):
    # Plot LCL, LFC, EL labels (if it fails, inform the user.)
    trans = transforms.blended_transform_factory(ax.transAxes, ax.transData)
    try:    
        ax.text(0.90, prof.mupcl.lclpres, '- LCL', verticalalignment='center', transform=trans, color='k', alpha=0.7)
    except:
        print("couldn't plot LCL")

    if np.isfinite(prof.mupcl.lfcpres):
        ax.text(0.90, prof.mupcl.lfcpres, '- LFC', verticalalignment='center', transform=trans, color='k', alpha=0.7)
    else:    
        print("couldn't plot LFC")

    try:
        ax.text(0.90, prof.mupcl.elpres, '- EL', verticalalignment='center', transform=trans, color='k', alpha=0.7)
    except:
        print("couldn't plot EL")

    return ax

def draw_heights(ax, prof):
    trans = transforms.blended_transform_factory(ax.transAxes,ax.transData)

    # Plot the height values on the skew-t, if there's an issue, inform the user.
    for h in [1000,2000,3000,4000,5000,6000,9000,12000,15000]:
        p = tab.interp.pres(prof, tab.interp.to_msl(prof, h))
        try:
            ax.text(0.01, p, str(h/1000) +' km -', verticalalignment='center', fontsize=9, transform=trans, color='r')
        except:
            print("problem plotting height label:", h)

    ax.text(0.01, prof.pres[prof.sfc], 'Sfc', verticalalignment='center', fontsize=9, transform=trans, color='r')
    return ax

def draw_effective_inflow_layer(ax, prof):
    # Plot the effective inflow layer on the Skew-T, like with the GUI (TODO: include the effective SRH on the top like in the GUI).
    trans = transforms.blended_transform_factory(ax.transAxes,ax.transData)
    ax.plot([0.2,0.3], [prof.ebottom, prof.ebottom], color='c', lw=2, transform=trans)
    ax.plot([0.25,0.25], [prof.etop, prof.ebottom], color='c', lw=2, transform=trans)
    ax.plot([0.2,0.3], [prof.etop, prof.etop], color='c', lw=2, transform=trans)

def draw_hodo_inset(ax, prof):
    # Draw the hodograph axes on the plot.
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes
    inset_axes = inset_axes(ax,width=1.7, # width = 30% of parent_bbox
                                        height=1.7, # height : 1 inch
                                        loc=1)
    inset_axes.get_xaxis().set_visible(False)
    inset_axes.get_yaxis().set_visible(False)

    # Draw the range rings around the hodograph.
    for i in range(10,90,10):
        circle = Circle((0,0),i,color='k',alpha=.3, fill=False)
        if i % 10 == 0 and i <= 50:
            inset_axes.text(-i,2,str(i), fontsize=8, horizontalalignment='center')
        inset_axes.add_artist(circle)
    inset_axes.set_xlim(-60,60)
    inset_axes.set_ylim(-60,60)
    inset_axes.axhline(y=0, color='k')
    inset_axes.axvline(x=0, color='k')
    #srwind = tab.params.bunkers_storm_motion(prof)

    return inset_axes

# Routine to plot the hodograph in segments (0-3 km, 3-6 km, etc.)
def plotHodo(axes, h, u, v, color='k'):
    for color, min_hght in zip(['r', 'g', 'b', 'k'], [3000,6000,9000,12000]):
        below_12km = np.where((h <= min_hght) & (h >= min_hght - 3000))[0]
        if len(below_12km) == 0:
            continue
        below_12km = np.append(below_12km, below_12km[-1] + 1)
        # Try except to ensure missing data doesn't cause this routine to fail.
        try:
            axes.plot(u[below_12km][~u.mask[below_12km]],v[below_12km][~v.mask[below_12km]], color +'-', lw=2)
        except:
            continue





''' Frequently used functions '''
from __future__ import division
import numpy as np
import numpy.ma as ma
from . constants import MISSING, TOL

__all__ = ['INT2STR','FLOAT2STR','MS2KTS', 'KTS2MS', 'MS2MPH']
__all__ += ['MPH2MS', 'MPH2KTS', 'KTS2MPH', 'M2FT', 'FT2M']
__all__ += ['IN2CM', 'CM2IN']
__all__ += ['vec2comp', 'comp2vec', 'mag', 'QC']

def INT2STR(val):
    '''
    Convert a value (float or int) into an integer string
    by rounding to the nearest integer and returning the 
    string. Used to handle case where value is nan or 
    masked.
    
    Parameters
    ----------
    val: float or int
    
    Returns
    -------
    Val rounded to the nearest int and converted to a string.
    
    '''
    if np.isnan(val):
        return '--'
    try:
        return str( int( np.round( val, 0 ) ) )
    except:
        return str(val)


def FLOAT2STR(val, precision):
    '''
    Convert a value (float or int) into a float string to the 
    decimal place of a specified precision. Used to handle the
    case where the value is nan or masked.
    
    Parameters
    ----------
    val: float or int
    precision: int
        used to specify the precision of the resulting rounded value
    
    Returns
    -------
    val rounded to the nearest specified decimal place and converted
    to a string.
    '''
    try:
        new_val = str( np.round( val, precision ) )
    except:
        new_val = str( val )

    if new_val.strip() == 'nan':
        return '--'
    else:
        return new_val

def MS2KTS(val):
    '''
    Convert meters per second to knots

    Parameters
    ----------
    val : float, numpy_array
        Speed (m/s)

    Returns
    -------
    Val converted to knots (float)

    '''
    return val * 1.94384449


def KTS2MS(val):
    '''
    Convert knots to meters per second

    Parameters
    ----------
    val : float, numpy_array
        Speed (kts)

    Returns
    -------
        Val converted to meters per second (float)

    '''
    return val * 0.514444


def MS2MPH(val):
    '''
    Convert meters per second to miles per hour

    Parameters
    ----------
    val : float, numpy_array
        Speed (m/s)

    Returns
    -------
    Val converted to miles per hour (float)

    '''
    return val * 2.23694


def MPH2MS(val):
    '''
    Convert miles per hour to meters per second

    Parameters
    ----------
    val : float, numpy_array
        Speed (mph)

    Returns
    -------
    Val converted to meters per second (float)

    '''
    return val * 0.44704


def MPH2KTS(val):
    '''
    Convert miles per hour to knots

    Parameters
    ----------
    val : float, numpy_array
        Speed (mph)

    Returns
    -------
    Val converted to knots (float)

    '''
    return val * 0.868976


def KTS2MPH(val):
    '''
    Convert knots to miles per hour

    Parameters
    ----------
    val : float, numpy_array
        Speed (kts)

    Returns
    -------
    Val converted to miles per hour (float)

    '''
    return val * 1.15078


def M2FT(val):
    '''
    Convert meters to feet

    Parameters
    ----------
    val : float, numpy_array
        Distance (m)

    Returns
    -------
        Val converted to feet (float)

    '''
    return val * 3.2808399


def FT2M(val):
    '''
    Convert feet to meters

    Parameters
    ----------
    val : float, numpy_array
        Distance (ft)

    Returns
    -------
        Val converted to meters (float)

    '''
    return val * 0.3048

def IN2CM(val):
    '''
    Convert inches to centimeters

    Parameters
    ----------
    val : float, numpy_array
          Distance (inches)

    Returns
    -------
        Val converted to centimeters (float)
    '''
    return val * 2.54

def CM2IN(val):
    '''
    Convert centimeters to inches

    Parameters
    ----------
    val : float, numpy_array
          Distance (centimeters)

    Returns
    -------
        Val converted to inches (float)
    '''
    return val / 2.54

def _vec2comp(wdir, wspd):
    '''
    Underlying function that converts a vector to its components

    Parameters
    ----------
    wdir : number, masked_array
        Angle in meteorological degrees
    wspd : number, masked_array
        Magnitudes of wind vector

    Returns
    -------
    u : number, masked_array (same as input)
        U-component of the wind
    v : number, masked_array (same as input)
        V-component of the wind

    '''
    u = wspd * ma.sin(np.radians(wdir)) * -1
    v = wspd * ma.cos(np.radians(wdir)) * -1
    return u, v


def vec2comp(wdir, wspd, missing=MISSING):
    '''
    Convert direction and magnitude into U, V components

    Parameters
    ----------
    wdir : number, array_like
        Angle in meteorological degrees
    wspd : number, array_like
        Magnitudes of wind vector (input units == output units)
    missing : number (optional)
        Optional missing parameter. If not given, assume default missing
        value from sharppy.sharptab.constants.MISSING

    Returns
    -------
    u : number, array_like (same as input)
        U-component of the wind (units are the same as those of input speed)
    v : number, array_like (same as input)
        V-component of the wind (units are the same as those of input speed)

    '''
    if not QC(wdir) or not QC(wspd):
        return ma.masked, ma.masked

    wdir = ma.asanyarray(wdir).astype(np.float64)
    wspd = ma.asanyarray(wspd).astype(np.float64)
    wdir.set_fill_value(missing)
    wspd.set_fill_value(missing)
    assert wdir.shape == wspd.shape, 'wdir and wspd have different shapes'
    if wdir.shape:
        wdir[wdir == missing] = ma.masked
        wspd[wspd == missing] = ma.masked
        wdir[wspd.mask] = ma.masked
        wspd[wdir.mask] = ma.masked
        u, v = _vec2comp(wdir, wspd)
        u[np.fabs(u) < TOL] = 0.
        v[np.fabs(v) < TOL] = 0.
    else:
        if wdir == missing:
            wdir = ma.masked
            wspd = ma.masked
        elif wspd == missing:
            wdir = ma.masked
            wspd = ma.masked
        u, v = _vec2comp(wdir, wspd)
        if ma.fabs(u) < TOL:
            u = 0.
        if ma.fabs(v) < TOL:
            v = 0.
    return u, v


def comp2vec(u, v, missing=MISSING):
    '''
    Convert U, V components into direction and magnitude

    Parameters
    ----------
    u : number, array_like
        U-component of the wind
    v : number, array_like
        V-component of the wind
    missing : number (optional)
        Optional missing parameter. If not given, assume default missing
        value from sharppy.sharptab.constants.MISSING

    Returns
    -------
    wdir : number, array_like (same as input)
        Angle in meteorological degrees
    wspd : number, array_like (same as input)
        Magnitudes of wind vector (input units == output units)

    '''
    if not QC(u) or not QC(v):
        return ma.masked, ma.masked

    u = ma.asanyarray(u).astype(np.float64)
    v = ma.asanyarray(v).astype(np.float64)
    u.set_fill_value(missing)
    v.set_fill_value(missing)
    wdir = np.degrees(np.arctan2(-u, -v))

    if wdir.shape:
        u[u == missing] = ma.masked
        v[v == missing] = ma.masked
        wdir[u.mask] = ma.masked
        wdir[v.mask] = ma.masked
        wdir[wdir < 0] += 360
        wdir[np.fabs(wdir) < TOL] = 0.
    else:
        if u == missing or v == missing:
            return ma.masked, ma.masked
        if wdir < 0:
            wdir += 360
        if np.fabs(wdir) < TOL:
            wdir = 0.
    return wdir, mag(u, v)


def mag(u, v, missing=MISSING):
    '''
    Compute the magnitude of a vector from its components

    Parameters
    ----------
    u : number, array_like
        U-component of the wind
    v : number, array_like
        V-component of the wind
    missing : number (optional)
        Optional missing parameter. If not given, assume default missing
        value from sharppy.sharptab.constants.MISSING

    Returns
    -------
    mag : number, array_like
        The magnitude of the vector (units are the same as input)

    '''
    if not QC(u) or not QC(v):
        return ma.masked

    u = np.ma.asanyarray(u).astype(np.float64)
    v = np.ma.asanyarray(v).astype(np.float64)
    u.set_fill_value(missing)
    v.set_fill_value(missing)
    if u.shape:
        u[u == missing] = ma.masked
        v[v == missing] = ma.masked
    else:
        if u == missing or v == missing:
            return ma.masked
    return ma.sqrt(u**2 + v**2)

def QC(val):
    '''
        Tests if a value is masked.
        
        '''
    if type(val) == type(ma.masked): return False
    return True





import numpy as np
from qtpy import QtGui, QtCore, QtWidgets
import sharppy.sharptab as tab
from sharppy.sharptab.constants import *
import platform

__all__ = ['backgroundSlinky', 'plotSlinky']

## Written by Greg Blumberg - CIMMS
## wblumberg@ou.edu

class backgroundSlinky(QtWidgets.QFrame):
    '''
    Draw the background frame and lines for the Storm Slinky.
    Draws onto a QPixmap.
    '''
    def __init__(self):
        super(backgroundSlinky, self).__init__()
        self.initUI()


    def initUI(self):
        ## window configuration settings,
        ## sich as padding, width, height, and
        ## min/max plot axes
        self.lpad = 5; self.rpad = 0
        self.tpad = 0; self.bpad = 20
        self.fpad = 5
        self.wid = self.size().width() - self.rpad
        self.hgt = self.size().height() - self.bpad
        self.tlx = self.rpad; self.tly = self.tpad
        self.brx = self.wid; self.bry = self.hgt
        ## center the frame on the slinky
        self.center_frame()
        if self.physicalDpiX() > 75:
            fsize = 7
        else:
            fsize = 9
        self.font_ratio = 0.0512
        self.title_font = QtGui.QFont('Helvetica', round(self.size().height() * self.font_ratio)+2)
        self.plot_font = QtGui.QFont('Helvetica', round(self.size().height() * self.font_ratio))
        self.title_metrics = QtGui.QFontMetrics( self.title_font )
        self.plot_metrics = QtGui.QFontMetrics( self.plot_font )
        self.os_mod = 0
        if platform.system() == "Windows":
            self.os_mod = self.plot_metrics.descent()

        ## get the pixel height of the font
        self.title_height = self.title_metrics.xHeight() + self.fpad
        self.plot_height = self.plot_metrics.xHeight() + self.fpad
        ## initialize the QPixmap that the frame and data will be drawn on
        self.plotBitMap = QtGui.QPixmap(self.width(), self.height())
        self.plotBitMap.fill(self.bg_color)
        ## plot the background
        self.plotBackground()  

    def center_frame(self):
        '''
        Center the slinky in the window.

        '''
        self.centerx = self.wid / 2; self.centery = self.hgt / 2
        self.mag = 7000.*1.7
        self.scale = (self.brx - self.tlx) / self.mag

    def resizeEvent(self, e):
        '''
        Handles the event the window is resized.
        '''
        self.initUI()

    def draw_frame(self, qp):
        '''
        Draw the background frame.
        
        Parameters
        ----------
        qp: QtGui.QPainter object
        
        '''
        ## set a new pen to draw with
        pen = QtGui.QPen(self.fg_color, 2, QtCore.Qt.SolidLine)
        qp.setPen(pen)
        qp.setFont(self.title_font)
        
        ## draw the borders in white
        qp.drawLine(self.tlx, self.tly, self.brx, self.tly)
        qp.drawLine(self.brx, self.tly, self.brx, self.bry)
        qp.drawLine(self.brx, self.bry, self.tlx, self.bry)
        qp.drawLine(self.tlx, self.bry, self.tlx, self.tly)

        yval = self.bry - self.title_height - self.os_mod - 2
        rect0 = QtCore.QRect(self.lpad, yval, 20, self.title_height)
        qp.drawText(rect0, QtCore.Qt.TextDontClip | QtCore.Qt.AlignLeft, 'Storm Slinky')
    
    def draw_axes(self, qp):
        '''
        Draw the X, Y Axes.
        
        Parameters
        ----------
        qp: QtGui.QPainter object

        '''
        ## initialize a white pen to draw the frame axes
        pen = QtGui.QPen(QtGui.QColor('#003366'), 2, QtCore.Qt.SolidLine)
        qp.setPen(pen)
        ## draw the frame axes
        qp.drawLine(self.centerx, self.tly, self.centerx, self.bry)
        qp.drawLine(self.tlx, self.centery, self.brx, self.centery)
    
    def plotBackground(self):
        '''
        Draws the background frame onto the QPixmap.
        '''
        qp = QtGui.QPainter()
        qp.begin(self.plotBitMap)
        qp.setRenderHint(qp.Antialiasing)
        qp.setRenderHint(qp.TextAntialiasing)
        ## draw the frame
        self.draw_axes(qp)
        self.draw_frame(qp)
        qp.end()

    def xy_to_pix(self, x, y):
        '''
        Function to convert (x, y) to pixel (xx, yy) coordinates.
        --------
        x: the x distance component
        y: the y distance component

        '''
        xx = self.centerx + (x * self.scale)
        yy = self.centery - (y * self.scale)
        return xx, yy


class plotSlinky(backgroundSlinky):
    '''
    Plots the data on the frame. Inherits from the
    backgroundSlinky class and draws on the QPixmap
    that is initialized there.
    '''
    def __init__(self, **kwargs):
        '''
        Initializes data variables needed to draw 
        the slinky by taking in a Profile object.
        
        Parameters
        ----------
        prof: a Profile Object
        
        '''
        self.bg_color = QtGui.QColor('#000000')
        self.fg_color = QtGui.QColor('#ffffff')
        self.use_left = False

        super(plotSlinky, self).__init__()
        self.prof = None
        self.pcl = None

        self.low_level_color = QtGui.QColor(RED)
        self.mid_level_color = QtGui.QColor("#00FF00")
        self.upper_level_color = QtGui.QColor(YELLOW)
        self.trop_level_color = QtGui.QColor("#00FFFF")

    def setProf(self, prof):
        self.prof = prof

        if self.use_left:
            self.smu = self.prof.srwind[2]
            self.smv = self.prof.srwind[3]
        else:
            self.smu = self.prof.srwind[0]
            self.smv = self.prof.srwind[1]

        self.pcl = None

        self.clearData()
        self.plotBackground()
        self.plotData()
        self.update()

    def setParcel(self, pcl):
        self.pcl = pcl

        if self.prof is not None:
            self.slinky_traj, self.updraft_tilt = tab.params.parcelTraj(self.prof, pcl, self.smu, self.smv)

        self.clearData()
        self.plotBackground()
        self.plotData()
        self.update()

    def setPreferences(self, update_gui=True, **prefs):
        self.bg_color = QtGui.QColor(prefs['bg_color'])
        self.fg_color = QtGui.QColor(prefs['fg_color'])

        self.low_level_color = QtGui.QColor(prefs['0_3_color'])
        self.mid_level_color = QtGui.QColor(prefs['3_6_color'])
        self.upper_level_color = QtGui.QColor(prefs['6_9_color'])
        self.trop_level_color = QtGui.QColor(prefs['9_12_color'])

        if update_gui:

            if self.use_left:
                self.smu = self.prof.srwind[2]
                self.smv = self.prof.srwind[3]
            else:
                self.smu = self.prof.srwind[0]
                self.smv = self.prof.srwind[1]

            self.slinky_traj, self.updraft_tilt = tab.params.parcelTraj(self.prof, self.pcl, self.smu, self.smv)

            self.clearData()
            self.plotBackground()
            self.plotData()
            self.update()

    def setDeviant(self, deviant):
        self.use_left = deviant == 'left'

        if self.use_left:
            self.smu = self.prof.srwind[2]
            self.smv = self.prof.srwind[3]
        else:
            self.smu = self.prof.srwind[0]
            self.smv = self.prof.srwind[1]

        self.slinky_traj, self.updraft_tilt = tab.params.parcelTraj(self.prof, self.pcl, self.smu, self.smv)

        self.clearData()
        self.plotBackground()
        self.plotData()
        self.update()

    def resizeEvent(self, e):
        '''
        Handles when the window is resized.
        
        Parameters
        ----------
        e: an Event object
        
        '''
        super(plotSlinky, self).resizeEvent(e)
        self.plotData()
    
    def paintEvent(self, e):
        '''
        Handles painting the QPixmap onto the frame.
        
        Parameters
        ----------
        e: an Event object
        
        '''
        ## this function handles painting the plot
        super(plotSlinky, self).paintEvent(e)
        ## create a new painter obkect
        qp = QtGui.QPainter()
        qp.begin(self)
        qp.drawPixmap(0,0,self.plotBitMap)
        qp.end()

    def plotData(self):
        '''
        Draws the data onto the QPixmap.
        '''
        if self.prof is None or self.pcl is None:
            return

        qp = QtGui.QPainter()
        qp.begin(self.plotBitMap)
        qp.setRenderHint(qp.Antialiasing)
        qp.setRenderHint(qp.TextAntialiasing)
        self.plotSlinky(qp)
        self.plotSMV(qp)
        self.plotTilt(qp)
        qp.end()

    def clearData(self):
        '''
        Handles the clearing of the pixmap
        in the frame.
        '''
        self.plotBitMap = QtGui.QPixmap(self.width(), self.height())
        self.plotBitMap.fill(self.bg_color)
   
    def plotSMV(self, qp):
        '''
        Draws the line representing the Storm Motion Vector.
        
        Parameters
        ----------
        qp: a QtGui.QPainter object
        
        '''
        ## set the pen
        pen = QtGui.QPen(self.fg_color, 2, QtCore.Qt.SolidLine)
        qp.setPen(pen)
        ## scale the vector to be visible in the window
        if tab.utils.QC(self.smu) and tab.utils.QC(self.smv):
            wdir, wspd = tab.utils.comp2vec(self.smu, self.smv)
            u, v = tab.utils.vec2comp(wdir, 3000)
            ## convert the unit space to pixel space
            motion_x, motion_y = self.xy_to_pix(u,v)
            center_x, center_y = self.xy_to_pix(0,0)
            qp.drawLine(motion_x,motion_y, center_x,center_y)

    def plotTilt(self, qp):
        '''
        Plots the data for the updraft tilt.
        
        Parameters
        ----------
        qp: a QtGui.QPainter object
        
        '''
        ## initialize a pen
        pen = QtGui.QPen(self.fg_color, 1, QtCore.Qt.SolidLine)
        qp.setPen(pen)
        qp.setFont(self.title_font)
        ## draw the text
        rect0 = QtCore.QRect(self.brx-30, self.tly+2, 30, self.title_height)
        qp.drawText(rect0, QtCore.Qt.TextDontClip | QtCore.Qt.AlignRight, tab.utils.INT2STR(self.updraft_tilt) + ' deg  ')

    def plotSlinky(self, qp):
        '''
        Plots the circles of the Storm Slinky.
        
        Parameters
        ----------
        qp: a QtGui.QPainter object
        
        '''
        ## set the pen
        pen = QtGui.QPen(self.trop_level_color, 1, QtCore.Qt.SolidLine)
        ## if there is no storm slinky, don't plot it!
        if self.slinky_traj is np.ma.masked:
            return

        has_el = self.pcl.bplus > 1e-3 and tab.utils.QC(self.pcl.elhght)

        ## loop through the parcel tradjectory in reverse
        for tradj in self.slinky_traj[:]:
            ## get the x, y, and z location of the updraft at each height
            x = tradj[0]
            y = tradj[1]
            z = tradj[2]

            if not tab.utils.QC(x) or not tab.utils.QC(y):
                continue

            ## set the various colors
            if has_el and z == self.slinky_traj[-1][2]:
                pen = QtGui.QPen(QtGui.QColor("#FF00FF"), 1, QtCore.Qt.SolidLine)
            elif z < 3000:
                pen = QtGui.QPen(self.low_level_color, 1, QtCore.Qt.SolidLine)
            elif z < 6000:
                pen = QtGui.QPen(self.mid_level_color, 1, QtCore.Qt.SolidLine)
            elif z < 9000:
                pen = QtGui.QPen(self.upper_level_color, 1, QtCore.Qt.SolidLine)
            elif z < 12000:
                pen = QtGui.QPen(self.trop_level_color, 1, QtCore.Qt.SolidLine)
            else:
                continue
            ## draw the circle
            qp.setPen(pen)
            xx, yy = self.xy_to_pix(x,y)
            center = QtCore.QPointF(xx, yy)
            qp.drawEllipse(center, 5, 5)

if __name__ == '__main__':
    app_frame = QtGui.QApplication([])    
    tester = plotSlinky()
    tester.setGeometry(100,100,121,138)
    tester.show()    
    app_frame.exec_()



from .. databases.sars import *
from .  import *
from .. io.spc_decoder import SPCDecoder
import numpy as np
from datetime import datetime
import os

def get_profile(fname, sars_type):
    # Create a convective profile object
    # fname - filename/SARS sounding string to load in
    # sars_type - string showing what SARS database (hail/supercell) to look for the raw file
    # Load in the data
    try:
        sars_fname = getSounding(fname[0].decode('utf-8'), sars_type)
    except:
        print("Unable to find data file for:", fname[0])
        return None
    dec = SPCDecoder(sars_fname)
    profs = dec.getProfiles()
    prof = profs._profs[''][0]
    dates = profs._dates
    prof.strictQC = True
    try:
        new_prof = profile.ConvectiveProfile.copy(prof)
    except Exception as e:
        print("There was a problem with the generation of the ConvectiveProfile:", str(e))
        return None

    return new_prof

def calc_inputs(new_prof, sars_type):
    # Grab the input values for SARS that were generated by SHARPpy in the ConvectiveProfile object
    # new_prof - the ConvectiveProfile object
    # sars_type - type of SARS (hail/supercell)
    sfc_6km_shear = utils.KTS2MS( utils.mag( new_prof.sfc_6km_shear[0], new_prof.sfc_6km_shear[1]) )
    sfc_3km_shear = utils.KTS2MS( utils.mag( new_prof.sfc_3km_shear[0], new_prof.sfc_3km_shear[1]) )
    sfc_9km_shear = utils.KTS2MS( utils.mag( new_prof.sfc_9km_shear[0], new_prof.sfc_9km_shear[1]) )
    h500t = interp.temp(new_prof, 500.)
    lapse_rate = params.lapse_rate( new_prof, 700., 500., pres=True )
    right_srh3km = new_prof.right_srh3km[0]
    right_srh1km = new_prof.right_srh1km[0]
    left_srh3km = new_prof.left_srh3km[0]
    left_srh1km = new_prof.left_srh1km[0]
    mucape = new_prof.mupcl.bplus
    mlcape = new_prof.mlpcl.bplus
    mllcl = new_prof.mlpcl.lclhght
    mumr = thermo.mixratio(new_prof.mupcl.pres, new_prof.mupcl.dwpc)
    #self.ship = params.ship(self)
    if sars_type == 'supercell':
        data = [mlcape, mllcl, h500t, lapse_rate, utils.MS2KTS(sfc_6km_shear), right_srh1km, utils.MS2KTS(sfc_3km_shear), utils.MS2KTS(sfc_9km_shear), 
                right_srh3km]
    else:
        data = [ mumr, mucape, h500t, lapse_rate, sfc_6km_shear,
                sfc_9km_shear, sfc_3km_shear, right_srh3km ]
    return np.round(np.asarray(data),1)
    
def check_supercell_cal(use_db=True):
    # Use to check the SARS supercell calibration
    database_fn = os.path.join( os.path.dirname( __file__ ), 'sars_supercell.txt' )
    supercell_db = np.loadtxt(database_fn, skiprows=1, dtype=bytes, comments="%%%%") 
    hits = 0
    fa = 0
    cn = 0
    miss = 0
    match = 0
    for f in supercell_db:
        mlcape = float(f[3])
        mllcl = float(f[5])
        h5temp = float(f[9])
        lr = float(f[11])
        shr = float(f[7])
        srh = float(f[6])
        srh3 = float(f[14])
        shr3 = float(f[12])
        shr9 = float(f[13])
        if use_db is True:
            out = supercell('sars_supercell.txt', mlcape, mllcl, h5temp, lr, shr, srh, shr3, shr9, srh3)
        else:
            new_prof = get_profile(f, 'supercell')
            if new_prof is None:
                continue
            out = new_prof.right_supercell_matches
        m = int(f[0].decode('utf-8') in out[0])
        if m == 0:
            data = calc_inputs(new_prof, 'supercell')
            #print("C:", data)
            #print("T:", [mlcape, mllcl, h5temp, lr, shr, srh, shr3, shr9, srh3])
        match += m
        #print(f[0], match)
        if out[-1] >= .5 and int(f[1]) > 0:
            hits += 1
        elif out[-1] >= .5 and int(f[1]) == 0:
            fa += 1
        elif out[-1] < .5 and int(f[1]) == 0:
            cn += 1
        elif out[-1] < .5 and int(f[1]) > 0:
            miss += 1
    print("--- SARS SUPERCELL CALIBRATION ---")
    print_stats(hits, cn, miss, fa, match)
    return {'hits': hits, 'cn': cn, 'miss':miss, 'fa': fa, 'match':match}

def print_stats(hits, cn, miss, fa, matches):
    # Print out the verification stats
    print("TOTAL SNDGS:", hits + cn + miss + fa )
    print("SELF MATCHES:", matches)
    print("HIT:", hits)
    print("MISS:", miss)
    print("FALSE ALARM:", fa)
    print("CORRECT NULL:", cn)
    print("ACCURACY: %.3f" % (float(hits+cn)/float(hits+cn+miss+fa)))
    print("BIAS: %.3f" % (float(hits+fa)/float(hits+miss)))
    print("POD: %.3f" % (float(hits)/float(hits+miss)))
    print("FAR: %.3f" % (float(fa)/float(fa+hits)))
    print("CSI: %.3f" % (float(hits)/float(hits + miss + fa)))
    print("TSS: %.3f" % (float(hits)/float(hits+miss) - float(fa)/float(fa+cn)))
    print()

def calc_verification(vals):
    stats = {}
    stats['num'] = vals['hits'] + vals['cn'] + vals['miss'] + vals['fa']
    for key in vals.keys():
        stats[key] = vals[key]
    hits = stats['hits']
    miss = stats['miss']
    fa = stats['fa']
    cn = stats['cn']
    stats["ACCURACY"] = float(hits+cn)/float(hits+cn+miss+fa)
    stats["BIAS"] = float(hits+fa)/float(hits+miss)
    stats["POD"] = float(hits)/float(hits+miss)
    stats["FAR"] = float(fa)/float(fa+hits)
    stats["CSI"] = float(hits)/float(hits + miss + fa)
    stats["TSS"] = float(hits)/float(hits+miss) - float(fa)/float(fa+cn)
    return stats

def check_hail_cal(use_db=True):
    # Check the calibration of the SARS hail 
    database_fn = os.path.join( os.path.dirname( __file__ ), 'sars_hail.txt' )
    hail_db = np.loadtxt(database_fn, skiprows=1, dtype=bytes)
    hits = 0
    cn = 0
    miss = 0
    fa = 0 
    match = 0
    for f in hail_db:
        mumr = float(f[4])
        mucape = float(f[3])
        lr = float(f[7])
        h5_temp = float(f[5])
        shr = float(f[10])
        shr9 = float(f[11])
        shr3 = float(f[9])
        srh = float(f[12])
        if use_db is True:
            out = hail('sars_hail.txt', mumr, mucape, h5_temp, lr, shr, shr9, shr3, srh)
        else:
            new_prof = get_profile(f, 'hail')
            if new_prof is None:
                continue
            out = new_prof.right_matches
        m = int(f[0].decode('utf-8') in out[0])
        if m == 0:
            data = calc_inputs(new_prof, 'hail')
            #print("C:", data)
            #print("T:", [mumr, mucape, h5_temp, lr, shr, shr9, shr3, srh])
        match += m
        #print(f[0], match)
        if out[-1] >= .5 and float(f[2]) >= 2:
            hits += 1
        elif out[-1] >= .5 and float(f[2]) < 2:
            fa += 1
        elif out[-1] < .5 and float(f[2]) < 2:
            cn += 1
        elif out[-1] < .5 and float(f[2]) >= 2:
            miss += 1
    print("--- SARS HAIL CALIBRATION ---")
    print_stats(hits, cn, miss, fa, match)
    return {'hits': hits, 'cn': cn, 'miss':miss, 'fa': fa, 'match':match}

#check_db = False
#check_supercell_cal(check_db)
#check_hail_cal(check_db)



import sys, os
import platform

import multiprocessing
try:
    import multiprocessing.forking as forking
except ImportError:
    if platform.system() == "Windows":
        import multiprocessing.popen_spawn_win32 as forking
    else:
        import multiprocessing.popen_fork as forking

_env_frozen = 'frozen'
_env_frozen_path = '_MEIPASS'
_env_mp_frozen_path = _env_frozen_path + '2'

def isFrozen():
    return getattr(sys, _env_frozen, False)
    
def frozenPath():
    return getattr(sys, _env_frozen_path, None)

def freezeSupport():
    if platform.system() == "Windows" and isFrozen():
        multiprocessing.freeze_support()

class _Popen(forking.Popen):
    def __init__(self, *args, **kw):
        if isFrozen():
            # We have to set original _MEIPASS2 value from sys._MEIPASS
            # to get --onefile mode working.
            # Last character is stripped in C-loader. We have to add
            # '/' or '\\' at the end.
            os.putenv(_env_mp_frozen_path, frozenPath() + os.sep)
        try:
            super(_Popen, self).__init__(*args, **kw)
        finally:
            if hasattr(sys, 'frozen'):
                # On some platforms (e.g. AIX) 'os.unsetenv()' is not
                # available. In those cases we cannot delete the variable
                # but only set it to the empty string. The bootloader
                # can handle this case.
                if hasattr(os, 'unsetenv'):
                    os.unsetenv(_env_mp_frozen_path)
                else:
                    os.putenv(_env_mp_frozen_path, '')


class Process(multiprocessing.Process):
    _Popen = _Popen

Queue = multiprocessing.Queue
